""" NIMPA: functions for neuro image processing and analysis
    including partial volume correction (PVC) and ROI extraction and analysis.
"""
__author__    = "Pawel Markiewicz"
__copyright__ = "Copyright 2018"
#-------------------------------------------------------------------------------

import numpy as np
import sys
import os
import platform
import shutil
import scipy.ndimage as ndi
import nibabel as nib

from collections import namedtuple
from subprocess import call
import datetime
import re
import multiprocessing

from pkg_resources import resource_filename

import imio
import regseg

import resources as rs

#> GPU routines only on Linux and Windows
if 'compute' in rs.CC_ARCH and platform.system() in ['Linux', 'Windows']:
    import improc

sitk_flag = True
try:
    import SimpleITK as sitk
except ImportError:
    sitk_flag = False

# possible extentions for DICOM files
dcmext = ('dcm', 'DCM', 'ima', 'IMA')

niiext = ('nii.gz', 'nii', 'img', 'hdr')

#>--------------------------
def num(s):
    ''' Converts the string to a float or integer number.
    '''
    try:
        return int(s)
    except ValueError:
        return float(s)
#>--------------------------

# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
# FUNCTIONS: T R I M   &   P A R T I A L   V O L U M E   E F F E C T S   A N D   C O R R E C T I O N
# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>

def trimim( fims,

            refim = '',
            affine=None,
            scale=2,
            divdim = 8**2,
            fmax = 0.05,
            
            int_order=0,
            outpath='',
            fname='',
            fcomment='',
            store_avg=False,
            store_img_intrmd=False,
            store_img=False,
            imdtype=np.float32,
            memlim=False,
            verbose=False):
    '''
    Trim and upsample PET image(s), e.g., for GPU execution,
    PVC correction, ROI sampling, etc.
    The input images 'fims' can be passed in multiple ways:
    1. as a string of the folder containing NIfTI files
    2. as a string of a NIfTI file path (this way a 4D image can be loaded).
    3. as a list of NIfTI file paths.
    4. as a 3D or 4D image
    Parameters:
    -----------
    refim:  Path to the reference image, which was already trimmed.
            Needs the trimming parameters stored in the NIfTI header in 'descrip'. 
    affine: affine matrix, 4x4, for all the input images in case when images
            passed as a matrix (all have to have the same shape and data type)
    scale: the scaling factor for upsampling has to be greater than 1
    divdim: image divisor, i.e., the dimensions are a multiple of this number (default 64)
    int_order: interpolation order (0-nearest neighbour, 1-linear, as in scipy)
    fmax: fraction of the max image value used for finding image borders for trimming
    outpath: output folder path
    fname: file name when image given as a numpy matrix
    fcomment: part of the name of the output file, left for the user as a comment
    store_img_intrmd: stores intermediate images with suffix '_i'
    store_avg: stores the average image (if multiple images are given)
    imdtype: data type for output images
    memlim: Ture for cases when memory is limited and takes more processing time instead.
    verbose: verbose mode [True/False]
    '''

    using_multiple_files = False

    # case when input folder is given
    if isinstance(fims, basestring) and os.path.isdir(fims):
        # list of input images (e.g., PET)
        fimlist = [os.path.join(fims,f) for f in os.listdir(fims) if f.endswith(niiext)]
        imdic = imio.niisort(fimlist, memlim=memlim)
        if not (imdic['N']>50 and memlim):
            imin = imdic['im']
        imshape = imdic['shape']
        affine = imdic['affine']
        fldrin = fims
        fnms = [os.path.basename(f).split('.nii')[0] for f in imdic['files'] if f!=None]
        # number of images/frames
        Nim = imdic['N']
        using_multiple_files = True

    # case when input file is a 3D or 4D NIfTI image
    elif isinstance(fims, basestring) and os.path.isfile(fims) and fims.endswith(niiext):
        imdic = imio.getnii(fims, output='all')
        imin = imdic['im']
        if imin.ndim==3:
            imin.shape = (1, imin.shape[0], imin.shape[1], imin.shape[2])
        imdtype = imdic['dtype']
        imshape = imdic['shape'][-3:]
        affine = imdic['affine']
        fldrin = os.path.dirname(fims)
        fnms = imin.shape[0] * [ os.path.basename(fims).split('.nii')[0] ]
        # number of images/frames
        Nim = imin.shape[0]

    # case when a list of input files is given
    elif isinstance(fims, list) and all([os.path.isfile(k) for k in fims]):
        imdic = imio.niisort(fims, memlim=memlim)
        if not (imdic['N']>50 and memlim):
            imin = imdic['im']
        imshape = imdic['shape']
        affine = imdic['affine']
        imdtype = imdic['dtype']
        fldrin = os.path.dirname(fims[0])
        fnms = [os.path.basename(f).split('.nii')[0] for f in imdic['files']]
        # number of images/frames
        Nim = imdic['N']
        using_multiple_files = True

    # case when an array [#frames, zdim, ydim, xdim].  Can be 3D or 4D
    elif isinstance(fims, (np.ndarray, np.generic)) and (fims.ndim==4 or fims.ndim==3):
        # check image affine
        if affine.shape!=(4,4):
            raise ValueError('Affine should be a 4x4 array.')
        # create a copy to avoid mutation when only one image (3D)
        imin = np.copy(fims)
        if fims.ndim==3:
            imin.shape = (1, imin.shape[0], imin.shape[1], imin.shape[2])
        imshape = imin.shape[-3:]
        fldrin = os.path.join(os.path.expanduser('~'), 'NIMPA_output')
        if fname=='':
            fnms = imin.shape[0] * [ 'NIMPA' ]
        else:
            fnms = imin.shape[0] * [ fname ]
        # number of images/frames
        Nim = imin.shape[0]

    else:
        raise TypeError('Wrong data type input.')


    #> if using reference trimmed image:
    ref_flag = False
    if os.path.isfile(refim):
        refdct = imio.getnii(refim, output='all')
        nii_descrp = refdct['hdr']['descrip'].item()
        if 'trim' in nii_descrp:
            #> find all the numbers (int and floats)
            parstr = re.findall(r'-*\d+\.*\d*', nii_descrp)
            ix0, ix1, iy0, iy1, iz0, scale, fmax = (num(s) for s in parstr)
            ref_flag = True
            print 'i> using the trimming parameters of the reference image.'
        else:
            print 'w> the reference image does not contain trimming info--using default.'

    

    #-------------------------------------------------------
    # store images in this folder
    if outpath=='':
        petudir = os.path.join(fldrin, 'trimmed')
    else:
        petudir = os.path.join(outpath, 'trimmed')
    imio.create_dir( petudir )
    #-------------------------------------------------------

    #-------------------------------------------------------
    # scale is preferred to be integer
    try:
        scale = int(scale)
    except ValueError:
        raise ValueError('e> scale has to be an integer.')
    # scale factor as the inverse of scale
    sf = 1/float(scale)
    if verbose:
        print 'i> upsampling scale {}, giving resolution scale factor {} for {} images.'.format(scale, sf, Nim)
    #-------------------------------------------------------

    #-------------------------------------------------------
    # scaled input image and get a sum image as the base for trimming
    if scale>1:
        newshape = (scale*imshape[0], scale*imshape[1], scale*imshape[2])
        imsum = np.zeros(newshape, dtype=imdtype)
        if not memlim:
            imscl = np.zeros((Nim,)+newshape, dtype=imdtype)
            for i in range(Nim):
                imscl[i,:,:,:] = ndi.interpolation.zoom(imin[i,:,:,:], (scale, scale, scale), order=int_order )
                imsum += imscl[i,:,:,:]
        else:
            for i in range(Nim):
                if Nim>50 and using_multiple_files:
                    imin_temp = imio.getnii(imdic['files'][i])
                    imsum += ndi.interpolation.zoom(imin_temp, (scale, scale, scale), order=int_order )
                    if verbose:
                        print 'i> image sum: read', imdic['files'][i]
                else:
                    imsum += ndi.interpolation.zoom(imin[i,:,:,:], (scale, scale, scale), order=int_order )
    else:
        imscl = imin
        imsum = np.sum(imin, axis=0)

    # smooth the sum image for improving trimming (if any)
    #imsum = ndi.filters.gaussian_filter(imsum, imio.fwhm2sig(4.0, voxsize=abs(affine[0,0])), mode='mirror')
    #-------------------------------------------------------
   
    if not ref_flag:
        # find the object bounding indexes in x, y and z axes, e.g., ix0-ix1 for the x axis
        qx = np.sum(imsum, axis=(0,1))
        ix0 = np.argmax( qx>(fmax*np.nanmax(qx)) )
        ix1 = ix0+np.argmin( qx[ix0:]>(fmax*np.nanmax(qx)) )

        qy = np.sum(imsum, axis=(0,2))
        iy0 = np.argmax( qy>(fmax*np.nanmax(qy)) )
        iy1 = iy0+np.argmin( qy[iy0:]>(fmax*np.nanmax(qy)) )

        qz = np.sum(imsum, axis=(1,2))
        iz0 = np.argmax( qz>(fmax*np.nanmax(qz)) )

        # find the maximum voxel range for x and y axes
        IX = ix1-ix0+1
        IY = iy1-iy0+1
        tmp = max(IX, IY)
        #> get the range such that it is divisible by 
        #> divdim (64 by default) for GPU execution
        IXY = divdim * ((tmp+divdim-1)/divdim)
        div = (IXY-IX)/2
        # x
        ix0 -= div
        ix1 += (IXY-IX)-div
        # y
        div = (IXY-IY)/2
        iy0 -= div
        iy1 += (IXY-IY)-div
        # z
        tmp = (len(qz)-iz0+1)
        IZ = divdim * ((tmp+divdim-1)/divdim)
        iz0 -= IZ-tmp+1
        
    # save the trimming parameters in a dic
    trimpar = {'x':(ix0, ix1), 'y':(iy0, iy1), 'z':(iz0), 'fmax':fmax}

    # new dims (z,y,x)
    newdims = (imsum.shape[0]-iz0, iy1-iy0+1, ix1-ix0+1)
    imtrim = np.zeros((Nim,)+newdims, dtype=imdtype)
    imsumt = np.zeros(newdims, dtype=imdtype)
    # in case of needed padding (negative indx resulting above)
    # the absolute values are supposed to work like padding in case the indx are negative
    iz0s, iy0s, ix0s = iz0, iy0, ix0
    iz0t, iy0t, ix0t = 0,0,0
    if iz0<0: 
        iz0s=0; 
        iz0t = abs(iz0)
        print '-----------------------------------------------------------------'
        print 'w> Correcting for trimming outside the original image (z-axis)'
        print '-----------------------------------------------------------------'
    if iy0<0: 
        iy0s=0; 
        iy0t = abs(iy0)
        print '-----------------------------------------------------------------'
        print 'w> Correcting for trimming outside the original image (y-axis)'
        print '-----------------------------------------------------------------'
    if ix0<0: 
        ix0s=0; 
        ix0t = abs(ix0)
        print '-----------------------------------------------------------------'
        print 'w> Correcting for trimming outside the original image (x-axis)'
        print '-----------------------------------------------------------------'
    

    #> in case the upper index goes beyond the scaled but untrimmed image
    iy1t = imsumt.shape[1]
    if iy1>=imsum.shape[1]:
        iy1t -= iy1+1

    #> the same for x
    ix1t = imsumt.shape[2]
    if ix1>=imsum.shape[2]:
        ix1t -= ix1+1


    # first trim the sum image
    imsumt[iz0t:, iy0t:iy1t, ix0t:ix1t] = imsum[iz0s:, iy0s:iy1+1, ix0s:ix1+1]

    #> new affine matrix for the scaled and trimmed image
    A = np.diag(sf*np.diag(affine))
    #> note half of new voxel offset is used for the new centre of voxels
    A[0,3] = affine[0,3] + A[0,0]*(ix0-0.5)        
    A[1,3] = affine[1,3] + (affine[1,1]*(imshape[1]-1) - A[1,1]*(iy1-0.5))
    A[2,3] = affine[2,3] - A[1,1]*0.5
    A[3,3] = 1


    # output dictionary
    dctout = {'affine': A, 'trimpar':trimpar, 'imsum':imsumt}

    # NIfTI image description (to be stored in the header)
    niidescr = 'trim(x,y,z):' \
               + str(trimpar['x'])+',' \
               + str(trimpar['y'])+',' \
               + str((trimpar['z'],)) \
               + ';scale='+str(scale) \
               + ';fmx='+str(fmax)
    
    # store the sum image
    if store_avg:
        fsum = os.path.join(petudir, 'avg_trimmed-upsampled-scale-'+str(scale)+fcomment+'.nii.gz')
        imio.array2nii( imsumt[::-1,::-1,:], A, fsum, descrip=niidescr)
        if verbose:  print 'i> saved averaged image to:', fsum
        dctout['fsum'] = fsum

    # list of file names for the upsampled and trimmed images
    fpetu = []
    # perform the trimming and save the intermediate images if requested
    for i in range(Nim):

        # memory saving option, second time doing interpolation
        if memlim:
            if Nim>50 and using_multiple_files:
                    imin_temp = imio.getnii(imdic['files'][i])
                    im = ndi.interpolation.zoom(imin_temp, (scale, scale, scale), order=int_order )
                    if verbose:
                        print 'i> image scaling:', imdic['files'][i]
            else:
                im = ndi.interpolation.zoom(imin[i,:,:,:], (scale, scale, scale), order=int_order )
        else:
            im = imscl[i,:,:,:]

        # trim the scaled image
        imtrim[i, iz0t:, iy0t:iy1t, ix0t:ix1t] = im[iz0s:, iy0s:iy1+1, ix0s:ix1+1]
        
        # save the up-sampled and trimmed PET images
        if store_img_intrmd:
            _frm = '_trmfrm'+str(i)
            _fstr = '_trimmed-upsampled-scale-'+str(scale) + _frm*(Nim>1) +fcomment
            fpetu.append( os.path.join(petudir, fnms[i]+_fstr+'_i.nii.gz') )
            imio.array2nii( imtrim[i,::-1,::-1,:], A, fpetu[i], descrip=niidescr)
            if verbose:  print 'i> saved upsampled PET image to:', fpetu[i]

    if store_img:
        _nfrm = '_nfrm'+str(Nim)
        fim = os.path.join(petudir, 'trimmed-upsampled-scale-'+str(scale))+_nfrm*(Nim>1)+fcomment+'.nii.gz'
        imio.array2nii( np.squeeze(imtrim[:,::-1,::-1,:]), A, fim, descrip=niidescr)
        dctout['fim'] = fim

    # file names (with paths) for the intermediate PET images
    dctout['fimi'] = fpetu
    dctout['im'] = np.squeeze(imtrim)
    dctout['N'] = Nim
    dctout['affine'] = A
        
    return dctout




# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
def psf_general(vx_size=(1,1,1), fwhm=(5, 5, 6), hradius=8, scale=2):
    '''
    Separable kernels for convolution executed on the GPU device
    The outputted kernels are in this order: z, y, x
    '''
    xSig = (scale*fwhm[0]/vx_size[0]) / (2*(2*np.log(2))**.5)
    ySig = (scale*fwhm[1]/vx_size[1]) / (2*(2*np.log(2))**.5)
    zSig = (scale*fwhm[2]/vx_size[2]) / (2*(2*np.log(2))**.5)

    # get the separated kernels
    x = np.arange(-hradius,hradius+1)
    xKrnl = np.exp(-(x**2/(2*xSig**2)))
    yKrnl = np.exp(-(x**2/(2*ySig**2)))
    zKrnl = np.exp(-(x**2/(2*zSig**2)))

    # normalise kernels
    xKrnl /= np.sum(xKrnl)
    yKrnl /= np.sum(yKrnl)
    zKrnl /= np.sum(zKrnl)

    krnl = np.array([zKrnl, yKrnl, xKrnl], dtype=np.float32)

    # for checking if the normalisation worked
    # np.prod( np.sum(krnl,axis=1) )

    # return all kernels together
    return krnl
# ------------------------------------------------------------------------------------------------------

def psf_measured(scanner='mmr', scale=1):
    if scanner=='mmr':
        # file name for the mMR's PSF and chosen scale
        fnm = 'PSF-17_scl-'+str(int(scale))+'.npy'
        fpth = resource_filename(__name__, '../../auxdata')
        fdat = os.path.join(fpth, fnm)
        # cdir = os.path.dirname(resource_filename(__name__, __file__))
        # niftypet_dir = os.path.dirname(os.path.dirname(cdir))
        # fdat = os.path.join( os.path.join(niftypet_dir , 'auxdata' ), fnm)

        # transaxial and axial PSF
        Hxy, Hz = np.load(fdat)
        krnl = np.array([Hz, Hxy, Hxy], dtype=np.float32)
    else:
        raise NameError('e> only Siemens mMR is currently supported.')
    return krnl
# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
def iyang(imgIn, krnl, imgSeg, Cnt, itr=5):
    '''partial volume correction using iterative Yang method
    imgIn: input image which is blurred due to the PSF of the scanner
    krnl: shift invariant kernel of the PSF
    imgSeg: segmentation into regions starting with 0 (e.g., background) and then next integer numbers
    itr: number of iteration (default 5)
    '''
    dim = imgIn.shape
    m = np.int32(np.max(imgSeg))
    m_a = np.zeros(( m+1, itr ), dtype=np.float32)

    for jr in range(0,m+1): 
        m_a[jr, 0] = np.mean( imgIn[imgSeg==jr] )

    # init output image
    imgOut = np.copy(imgIn)

    # iterative Yang algorithm:
    for i in range(0, itr):
        if Cnt['VERBOSE']: print 'i> PVC Yang iteration =', i
        # piece-wise constant image
        imgPWC = imgOut
        imgPWC[imgPWC<0] = 0
        for jr in range(0,m+1):
            imgPWC[imgSeg==jr] = np.mean( imgPWC[imgSeg==jr] )
        
        #> blur the piece-wise constant image using either:
        #> (1) GPU convolution with a separable kernel (x,y,z), or
        #> (2) CPU, Python-based convolution
        if 'CCARCH' in Cnt and 'compute' in Cnt['CCARCH']:

            #> convert to dimensions of GPU processing [y,x,z]
            imin_d = np.transpose(imgPWC, (1, 2, 0))
            imout_d = np.zeros(imin_d.shape, dtype=np.float32)
            improc.convolve(imout_d, imin_d, krnl, Cnt)
            imgSmo = np.transpose(imout_d, (2,0,1))
        else:
            hxy = np.outer(krnl[1,:], krnl[2,:])
            hxyz = np.multiply.outer(krnl[0,:], hxy)
            imgSmo = ndi.convolve(imgPWC, hxyz, mode='constant', cval=0.)
        
        # correction factors
        imgCrr = np.ones(dim, dtype=np.float32)
        imgCrr[imgSmo>0] = imgPWC[imgSmo>0] / imgSmo[imgSmo>0]
        imgOut = imgIn * imgCrr;
        for jr in range(0,m+1):
            m_a[jr, i] = np.mean( imgOut[imgSeg==jr] )

    return imgOut, m_a
# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>


# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
# G E T   P A R C E L L A T I O N S   F O R   P V C   A N D   R O I   E X T R A C T I O N
# ------------------------------------------------------------------------------------------------------
def pvc_iyang(
        petin,
        mridct,
        Cnt,
        pvcroi,
        krnl,
        itr=5,
        tool='niftyreg',
        faff='',
        outpath='',
        fcomment='',
        store_img=False,
        store_rois=False,
        matlab_eng = ''
    ):
    ''' Perform partial volume (PVC) correction of PET data (petin) using MRI data (mridct).
        The PVC method uses iterative Yang method.
        GPU based convolution is the key routine of the PVC. 
        Input:
        -------
        petin:  either a dictionary containing image data, file name and affine transform,
                or a string of the path to the NIfTI file of the PET data.
        mridct: a dictionary of MRI data, including the T1w image, which can be given
                in DICOM (field 'T1DCM') or NIfTI (field 'T1nii').  The T1w image data
                is needed for co-registration to PET if affine is not given in the text
                file with its path in faff.
        Cnt:    a dictionary of paths for third-party tools:
                * dcm2niix: Cnt['DCM2NIIX']
                * niftyreg, resample: Cnt['RESPATH']
                * niftyreg, rigid-reg: Cnt['REGPATH']
                * verbose mode on/off: Cnt['VERBOSE'] = True/False
        pvcroi: list of regions (also a list) with number label to distinguish
                the parcellations.  The numbers correspond to the image values
                of the parcellated T1w image.  E.g.:
                pvcroi = [
                    [36], # ROI 1 (single parcellation region)
                    [35], # ROI 2
                    [39, 40, 72, 73, 74], # ROI 3 (region consisting of multiple parcellation regions)
                    ...
                ]
        kernel: the point spread function (PSF) specific for the camera and the object.  
                It is given as a 3x17 matrix, a 17-element kernel for each dimension (x,y,z).
                It is used in the GPU-based convolution using separable kernels.
        outpath:path to the output of the resulting PVC images
        faff:   a text file of the affine transformations needed to get the MRI into PET space.
                If not provided, it will be obtained from the performed rigid transformation.
                For this the MR T1w image is required.  If faff and T1w image are not provided,
                it will results in exception/error.
        fcomment:a string used in naming the produced files, helpful for distinguishing them.
        tool:   co-registration tool.  By default it is NiftyReg, but SPM is also 
                possible (needs Matlab engine and more validation)
        itr:    number of iterations used by the PVC.  5-10 should be enough (5 default)
    '''

    # get all the input image properties
    if isinstance(petin, dict):
        im = imdic['im']
        fpet = imdic['fpet']
        B = imdic['affine']
    elif isinstance(petin, basestring) and os.path.isfile(petin):
        imdct = imio.getnii(petin, output='all')
        im = imdct['im']
        B = imdct['affine']
        fpet = petin
    
    if im.ndim!=3:
        raise IndexError('Only 3D images are expected in this method of partial volume correction.')

    # check if brain parcellations exist in NIfTI format
    if not os.path.isfile(mridct['T1lbl']):
        raise NameError('MissingLabels')

    #> output dictionary
    outdct = {}

    # establish the output folder
    if outpath=='':        
        prcl_dir = os.path.dirname(mridct['T1lbl'])
        pvc_dir =  os.path.join(os.path.dirname(fpet), 'PVC')
    else:
        prcl_dir = outpath
        pvc_dir =  os.path.join(os.path.dirname(fpet), 'PVC')

    #> create folders
    imio.create_dir(prcl_dir)
    imio.create_dir(pvc_dir)
    if store_rois:
        rois_dir = os.path.join(pvc_dir, 'ROIs')
        imio.create_dir(rois_dir)

    #==================================================================
    #if affine transf. (faff) is given then take the T1 and resample it too.
    if isinstance(faff, basestring) and not os.path.isfile(faff):
        # faff is not given; get it by running the affine; get T1w to PET space

        ft1w = imio.pick_t1w(mridct)

        if tool=='spm':
            regdct = regseg.coreg_spm(
                fpet,
                ft1w,
                fcomment = fcomment,
                outpath=os.path.join(outpath,'PET', 'positioning')
            )
            #> save Matlab engine for later resampling
            matlab_eng = regdct['matlab_eng']

        elif tool=='niftyreg':
            regdct = regseg.affine_niftyreg(
                fpet,
                ft1w,
                outpath=os.path.join(outpath,'PET', 'positioning'),
                fcomment = fcomment,
                executable = Cnt['REGPATH'],
                omp = multiprocessing.cpu_count()/2,
                rigOnly = True,
                affDirect = False,
                maxit=5,
                speed=True,
                pi=50, pv=50,
                smof=0, smor=0,
                rmsk=True,
                fmsk=True,
                rfwhm=15., #millilitres
                rthrsh=0.05,
                ffwhm = 15., #millilitres
                fthrsh=0.05,
                verbose=Cnt['VERBOSE']
            )
        faff = regdct['faff']
    else:
        if tool=='spm' and matlab_eng == '':
            print 'i> starting Matlab for SPM...'
            import matlab.engine
            matlab_eng = matlab.engine.start_matlab()


    # resample the GIF T1/labels to upsampled PET
    # file name of the parcellation (GIF-based) upsampled to PET
    fgt1u = os.path.join(
            prcl_dir,
            os.path.basename(mridct['T1lbl']).split('.')[0]\
            +'_registered_trimmed'+fcomment+'.nii.gz')
        
    if tool=='niftyreg':
        if os.path.isfile( Cnt['RESPATH'] ):
            cmd = [Cnt['RESPATH'],  '-ref', fpet,  '-flo', mridct['T1lbl'],
                   '-trans', faff, '-res', fgt1u, '-inter', '0']
            if not Cnt['VERBOSE']: cmd.append('-voff')
            call(cmd)
        else:
            raise IOError('e> path to resampling executable is incorrect!')
        
    elif tool=='spm':
        fout = regseg.resample_spm(
                    fpet,
                    mridct['T1lbl'],
                    faff,
                    fimout = fgt1u,
                    matlab_eng = matlab_eng,
                    intrp = 0.,
                    del_ref_uncmpr = True,
                    del_flo_uncmpr = True,
                    del_out_uncmpr = True,
            )

    #==================================================================

    # Get the parcellation labels in the upsampled PET space
    dprcl = imio.getnii(fgt1u, output='all')
    prcl = dprcl['im']

    #> path to parcellations in NIfTI format
    prcl_pth = os.path.split(mridct['T1lbl'])

    #---------------------------------------------------------------------------
    #> get the parcellation specific for PVC based on the current parcellations
    imgroi = prcl.copy();  imgroi[:] = 0
    
    #> number of segments, without the background
    nSeg = len(pvcroi)
    
    #> create the image of numbered parcellations
    for k in range(nSeg):
        for m in pvcroi[k]:
            imgroi[prcl==m] = k+1

    #> save the PCV ROIs to a new NIfTI file
    if store_rois:
        froi = os.path.join(rois_dir, prcl_pth[1].split('.nii')[0]+'_PVC-ROIs-inPET.nii.gz')
        imio.array2nii(
            imgroi,
            dprcl['affine'],
            froi,
            trnsp = (dprcl['transpose'].index(0),
                     dprcl['transpose'].index(1),
                     dprcl['transpose'].index(2)),
            flip = dprcl['flip'])
        outdct['froi'] = froi
    #---------------------------------------------------------------------------


    #---------------------------------------------------------------------------
    # run iterative Yang PVC
    imgpvc, m_a = iyang(im, krnl, imgroi, Cnt, itr=itr)
    #---------------------------------------------------------------------------

    outdct['im'] = imgpvc
    outdct['imroi'] = imgroi
    outdct['faff'] = faff

    if store_img:
        fpvc = os.path.join( pvc_dir,
                             os.path.split(fpet)[1].split('.nii')[0]+'_PVC'+fcomment+'.nii.gz')
        imio.array2nii( imgpvc[::-1,::-1,:], B, fpvc, descrip='pvc=iY')
        outdct['fpet'] = fpvc

    return outdct


#==============================================================
# Convert CT units (HU) to PET mu-values
def ct2mu(im):
    '''HU units to 511keV PET mu-values
        https://link.springer.com/content/pdf/10.1007%2Fs00259-002-0796-3.pdf
        C. Burger, et al., PET attenuation coefficients from CT images, 
    '''

    # convert nans to -1024 for the HU values only
    im[np.isnan(im)] = -1024
    # constants
    muwater  = 0.096
    mubone   = 0.172
    rhowater = 0.184
    rhobone  = 0.428
    uim = np.zeros(im.shape, dtype=np.float32)
    uim[im<=0] = muwater * ( 1+im[im<=0]*1e-3 )
    uim[im> 0] = muwater+im[im>0]*(rhowater*(mubone-muwater)/(1e3*(rhobone-rhowater)))
    # remove negative values
    uim[uim<0] = 0
    return uim
#==============================================================



#===============================================================================
def centre_mass_img(imdct):
    ''' Calculate the centre of mass of an image along each axes (x,y,z),
        separately.
        Arguments:
        imdct - the image dictionary with the image and header data.
        Output the list of the centre of mass for each axis.
    '''
    
    #> initialise centre of mass array
    com = np.zeros(3, dtype=np.float32)
    #> total image sum
    imsum = np.sum(imdct['im'])

    for ind_ax in [-1, -2, -3]:
        #> list of axes
        axs = range(imdct['im'].ndim)
        del axs[ind_ax]
        #> indexed centre of mass
        icom = np.sum( np.sum(imdct['im'], axis=tuple(axs)) \
                * np.arange(imdct['im'].shape[ind_ax]))/imsum
        #> centre of mass in mm
        com[abs(ind_ax)-1] = icom * imdct['hdr']['pixdim'][abs(ind_ax)]

    return com
#===============================================================================


#===============================================================================
def nii_modify(
        nii,
        fimout = '',
        outpath = '',
        fcomment = '',
        voxel_range=[]):

    ''' Modify the NIfTI image given either as a file path or a dictionary,
        obtained by nimpa.getnii(file_path).
    '''

    if isinstance(nii, basestring) and os.path.isfile(nii):
        dctnii = imio.getnii(nii, output='all')
        fnii = nii
    if isinstance(nii, dict) and 'im' in nii:
        dctnii = nii
        if 'fim' in dctnii:
            fnii = dctnii['fim']

    #---------------------------------------------------------------------------
    #> output path
    if outpath=='' and fimout!='' and '/' in fimout:
        opth = os.path.dirname(fimout)
        if opth=='' and isinstance(fnii, basestring) and os.path.isfile(fnii):
            opth = os.path.dirname(nii)
        fimout = os.path.basename(fimout)

    elif outpath=='' and isinstance(fnii, basestring) and os.path.isfile(fnii):
        opth = os.path.dirname(fnii)
    else:
        opth = outpath
    imio.create_dir(opth)

    #> output floating and affine file names
    if fimout=='':

        if fcomment=='':
            fcomment += '_nimpa-modified'

        fout = os.path.join(
                opth,
                os.path.basename(fnii).split('.nii')[0]+fcomment+'.nii.gz')
    else:
        fout = os.path.join(
                opth,
                fimout.split('.nii')[0]+'.nii.gz')
    #---------------------------------------------------------------------------


    #> reduce the max value to 255
    if voxel_range and len(voxel_range)==1:
        im = voxel_range[0] * dctnii['im']/np.max(dctnii['im'])
    elif voxel_range and len(voxel_range)==2:
        #> normalise into range 0-1
        im = (dctnii['im'] - np.min(dctnii['im'])) / np.ptp(dctnii['im'])
        #> convert to voxel_range 
        im = voxel_range[0] + im*(voxel_range[1] - voxel_range[0])
    else:
        return None

    #> output file name for the extra reference image
    imio.array2nii(
        im,
        dctnii['affine'],
        fout,
        trnsp = (dctnii['transpose'].index(0),
                 dctnii['transpose'].index(1),
                 dctnii['transpose'].index(2)),
        flip = dctnii['flip'])

    return {'fim':fout, 'im':im, 'affine':dctnii['affine']}





#===============================================================================





#  ____________________________________________________________________________
# |                                                                            |
# |                 M R   B I A S   C O R R E C T I O N                        |
# |____________________________________________________________________________|

def bias_field_correction(
        fmr,
        fimout = '',
        outpath = '',
        fcomment = '_N4bias',
        executable = '',
        exe_options = [],
        sitk_image_mask = True,
        verbose = False,):

    ''' Correct for bias field in MR image(s) given in <fmr> as a string
        (single file) or as a list of strings (multiple files).  

        Output dictionary with the bias corrected file names.

        Options:
        - fimout:       The name (with path) of the output file.  It's 
                        ignored when multiple files are given as input.  If
                        given for a single file name, the <outpath> and
                        <fcomment> options are ignored.
        - outpath:      Path to the output folder
        - fcomment:     A prefix comment to the file name
        - executable:   The path to the executable, overrides the above
                        choice of software;  if 'sitk' is given instead
                        of the path, the Python module SimpleITK will be 
                        used if it is available.
        - exe_options:  Options for the executable in the form of a list of
                        strings.
        - sitk_image_mask:  Image masking will be used if SimpleITK is
                            chosen.
    '''

    if executable=='sitk' and not 'SimpleITK' in sys.modules:
        print  'e> if SimpleITK module is required for bias correction' \
                +' it needs to be first installed.\n' \
                +'   Install the module by:\n' \
                +'   conda install -c https://conda.anaconda.org/simpleitk SimpleITK=1.2.0\n' \
                +'   or pip install SimpleITK'
        return None

    #---------------------------------------------------------------------------
    # INPUT
    #---------------------------------------------------------------------------
    #> path to a single file 
    if isinstance(fmr, basestring) and os.path.isfile(fmr):
        fins = [fmr]

    #> list of file paths
    elif isinstance(fmr, list) and all([os.path.isfile(f) for f in fmr]):
        fins = fmr
        print 'i> multiple input files => ignoring the single output file name.'
        fimout = ''

    #> path to a folder
    elif isinstance(fmr, basestring) and os.path.isdir(fmr):
        fins = [os.path.join(fmr, f) for f in os.listdir(fmr) if f.endswith(('.nii', '.nii.gz'))]
        print 'i> multiple input files from input folder.'
        fimout = ''

    else:
        raise ValueError('could not decode the input of floating images.')
    #---------------------------------------------------------------------------




    #---------------------------------------------------------------------------
    # OUTPUT
    #---------------------------------------------------------------------------
    #> output path
    if outpath=='' and fimout!='':
        opth = os.path.dirname(fimout)
        if opth=='':
            opth = os.path.dirname(fmr)
            fimout = os.path.join(opth, fimout)
        n4opth = opth
        fcomment = ''

    elif outpath=='':
        opth = os.path.dirname(fmr)
        #> N4 bias correction specific folder
        n4opth = os.path.join(opth, 'N4bias')

    else:
        opth = outpath
        #> N4 bias correction specific folder
        n4opth = os.path.join(opth, 'N4bias')

    imio.create_dir(n4opth)

    outdct = {}
    #---------------------------------------------------------------------------



    for fin in fins:

        if verbose:
            print 'i> input for bias correction:\n', fin
        
        if fimout=='':
            # split path
            fspl = os.path.split(fin)
            
            # N4 bias correction file output paths
            fn4 = os.path.join( n4opth, fspl[1].split('.nii')[0]\
                                + fcomment +'.nii.gz')
        else:
            fn4 = fimout

        if not os.path.exists(fn4):

            if executable=='sitk':
                #==============================================
                # SimpleITK Bias field correction for T1 and T2
                #==============================================
                #> initialise the corrector
                corrector = sitk.N4BiasFieldCorrectionImageFilter()
                # numberFilltingLevels = 4

                # read input file        
                im = sitk.ReadImage(fin)

                #> create a object specific mask
                fmsk = os.path.join( n4opth, fspl[1].split('.nii')[0] +'_sitk_mask.nii.gz')
                msk = sitk.OtsuThreshold(im, 0, 1, 200)
                sitk.WriteImage(msk, fmsk)
                
                #> cast to 32-bit float
                im = sitk.Cast( im, sitk.sitkFloat32 )

                #-------------------------------------------
                print 'i> correcting bias field for', fin 
                n4out = corrector.Execute(im, msk)
                sitk.WriteImage(n4out, fn4)
                #-------------------------------------------

                if sitk_image_mask:
                    if not 'fmsk' in outdct: outdct['fmsk'] = []
                    outdct['fmsk'].append(fmsk)


            elif os.path.basename(executable)=='N4BiasFieldCorrection' \
                    and os.path.isfile(executable):
                
                cmd = [executable,  '-i', fin, '-o', fn4]
            
                if verbose and os.path.basename(executable)=='N4BiasFieldCorrection':
                    cmd.extend(['-v', '1'])

                cmd.extend(exe_options)
                
                call(cmd)

                if 'command' not in outdct:
                    outdct['command'] = []
                outdct['command'].append(cmd)

            elif os.path.isfile(executable):

                cmd = [executable]
                cmd.extend(exe_options)
                call(cmd)

                if 'command' not in outdct:
                    outdct['command'] = cmd
        
        else:
            print '   N4 bias corrected file seems already existing.'

        #> output to dictionary 
        if not 'fim' in outdct: outdct['fim'] = []
        outdct['fim'].append(fn4)


    return outdct


#-------------------------------------------------------------------------------------
def pet2pet_rigid(fref, fflo, Cnt, outpath='', rmsk=True, rfwhm=15., rthrsh=0.05, pi=50, pv=50, smof=0, smor=0):

    #create a folder for PET images registered to ref PET
    if outpath=='':
        outpath = os.path.dirname(fflo)
    odir = os.path.join(outpath,'PET2PET')
    imio.create_dir(odir)

    if rmsk:
        fimdir = os.path.join(odir, 'tmp')
        imio.create_dir(fimdir)
        fmsk = os.path.join(fimdir, 'rmask.nii.gz')
        imdct = nimpa.getnii(fref, output='all')
        smoim = ndi.filters.gaussian_filter(imdct['im'],
                                            imio.fwhm2sig(rfwhm, voxsize=imdct['affine'][0,0]), mode='mirror')
        thrsh = rthrsh*smoim.max()
        immsk = np.int8(smoim>thrsh)
        for iz in range(immsk.shape[0]):
            for iy in range(immsk.shape[1]):
                ix0 = np.argmax(immsk[iz,iy,:]>0)
                ix1 = immsk.shape[2] - np.argmax(immsk[iz,iy,::-1]>0)
                if (ix1-ix0) > immsk.shape[2]-10: continue
                immsk[iz,iy,ix0:ix1] = 1
        imio.array2nii( immsk[::-1,::-1,:], imio.getnii_affine(fref), fmsk)

    # output in register with ref PET
    fout = os.path.join(odir, 'PET-r-to-'+os.path.basename(fref).split('.')[0]+'.nii.gz')
    # text file for the affine transform T1w->PET
    faff   = os.path.join(odir, 'affine-PET-r-to-'+os.path.basename(fref).split('.')[0]+'.txt')  
    # call the registration routine
    if os.path.isfile( Cnt['REGPATH'] ):
        cmd = [Cnt['REGPATH'],
             '-ref', fref,
             '-flo', fflo,
             '-rigOnly', '-speeeeed',
             '-aff', faff,
             '-pi', str(pi),
             '-pv', str(pv),
             '-smooF', str(smof),
             '-smooR', str(smor),
             '-res', fout]
        if rmsk: 
            cmd.append('-rmask')
            cmd.append(fmsk)
        if not Cnt['VERBOSE']: cmd.append('-voff')
        print cmd
        call(cmd)
    else:
        print 'e> path to registration executable is incorrect!'
        raise StandardError('No registration executable found')
        
    return faff, fout
# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>


def mr2pet_rigid(
        fpet, mridct, Cnt,
        outpath='',
        fcomment='',
        rmsk=True,
        rfwhm=15.,
        rthrsh=0.05,
        fmsk=True,
        ffwhm = 15.,
        fthrsh=0.05,
        pi=50, pv=50,
        smof=0, smor=0):

    # create output path if given
    if outpath!='':
        imio.create_dir(outpath)

    # --- MR T1w
    if 'T1nii' in mridct and os.path.isfile(mridct['T1nii']):
        ft1w = mridct['T1nii']
    elif 'T1bc' in mridct and os.path.isfile(mridct['T1bc']):
        ft1w = mridct['T1bc']
    elif 'T1DCM' in mridct and os.path.exists(mridct['MRT1W']):
        # create file name for the converted NIfTI image
        fnii = 'converted'
        call( [ Cnt['DCM2NIIX'], '-f', fnii, mridct['T1nii'] ] )
        ft1nii = glob.glob( os.path.join(mridct['T1nii'], '*converted*.nii*') )
        ft1w = ft1nii[0]
    else:
        print 'e> disaster: no T1w image!'
        sys.exit()

    #create a folder for MR images registered to PET
    if outpath!='':
        mrodir = os.path.join(outpath,'T1w2PET')
        fimdir = os.path.join(outpath, os.path.join('T1w2PET','tmp'))
    else:
        mrodir = os.path.join(os.path.dirname(ft1w),'mr2pet')
        fimdir = os.path.join(os.path.basename(ft1w), 'tmp')
    imio.create_dir(mrodir)
    imio.create_dir(fimdir)

    if rmsk:
        f_rmsk = os.path.join(fimdir, 'rmask.nii.gz')
        imdct = imio.getnii(fpet, output='all')
        smoim = ndi.filters.gaussian_filter(imdct['im'],
                                            imio.fwhm2sig(rfwhm, voxsize=abs(imdct['affine'][0,0])), mode='mirror')
        thrsh = rthrsh*smoim.max()
        immsk = np.int8(smoim>thrsh)
        immsk = regseg.imfill(immsk)
        imio.array2nii( immsk[::-1,::-1,:], imdct['affine'], f_rmsk)
    if fmsk:
        f_fmsk = os.path.join(fimdir, 'fmask.nii.gz')
        imdct = imio.getnii(ft1w, output='all')
        smoim = ndi.filters.gaussian_filter(imdct['im'],
                                            imio.fwhm2sig(ffwhm, voxsize=abs(imdct['affine'][0,0])), mode='mirror')
        thrsh = fthrsh*smoim.max()
        immsk = np.int8(smoim>thrsh)
        immsk = regseg.imfill(immsk)
        imio.array2nii( immsk[::-1,::-1,:], imdct['affine'], f_fmsk)

    # if provided, separate the comment with underscore
    if fcomment!='': fcomment = '_'+fcomment
    # output for the T1w in register with PET
    ft1out = os.path.join(mrodir, 'T1w-r-to-'+os.path.basename(fpet).split('.')[0]+fcomment+'.nii.gz')
    # text file for the affine transform T1w->PET
    faff   = os.path.join(mrodir, 'affine-T1w-r-to-'+os.path.basename(fpet).split('.')[0]+fcomment+'.txt')  
    # call the registration routine
    if os.path.isfile( Cnt['REGPATH'] ):
        cmd = [Cnt['REGPATH'],
             '-ref', fpet,
             '-flo', ft1w,
             '-rigOnly', '-speeeeed',
             '-aff', faff,
             '-pi', str(pi),
             '-pv', str(pv),
             '-smooF', str(smof),
             '-smooR', str(smor),
             '-res', ft1out]
        if rmsk: 
            cmd.append('-rmask')
            cmd.append(f_rmsk)
        if fmsk:
            cmd.append('-fmask')
            cmd.append(f_fmsk)
        if 'VERBOSE' in Cnt and not Cnt['VERBOSE']: cmd.append('-voff')
        print cmd
        call(cmd)
    else:
        print 'e> path to registration executable is incorrect!'
        sys.exit()
        
    return faff




# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
# OUTDATED
# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>

def roi_extraction(imdic, amyroi, datain, Cnt, use_stored=False):
    '''
    Extracting ROI values of upsampled PET.  If not provided it will be upsampled.
    imdic: dictionary of all image parts (image, path to file and affine transformation)    
     > fpet: file name for quantitative PET
     > faff: file name and path to the affine (rigid) transformations, most likely for upscaled PET
    amyroi: module with ROI definitions in numbers
    datain: dictionary of all the input data
    '''

    fpet = imdic['fpet']
    faff = imdic['faff']
    im   = imdic['im']

    if Cnt['VERBOSE']: print 'i> extracting ROI values'

    # split the labels path
    lbpth = os.path.split(datain['T1lbl'])
    # # MR T1w NIfTI
    # if os.path.isfile(datain['T1nii']):
    #     ft1w = datain['T1nii']
    # elif os.path.isfile(datain['T1bc']):
    #     ft1w = datain['T1nii']
    # else:
    #     print 'e> no T1w NIfTI!'
    #     sys.exit()

    prcl_dir = os.path.dirname(datain['T1lbl'])

    dscrp = imio.getnii_descr(fpet)
    impvc = False
    if 'pvc' in dscrp.keys():       
        impvc = True
        if Cnt['VERBOSE']: print 'i> the provided image is partial volume corrected.'

    #------------------------------------------------------------------------------------------------------------  
    # next steps after having sorted out the upsampled input image and the rigid transformation for the T1w -> PET:

    # Get the labels before resampling to PET space so that the regions can be separated (more difficult to separate after resampling)
    if os.path.isfile(datain['T1lbl']):
        nilb = nib.load(datain['T1lbl'])
        A = nilb.get_sform()
        imlb = nilb.get_data()
    else:
        print 'e> parcellation label image not present!'
        sys.exit()

    # ===============================================================
    # get the segmentation/parcellation by creating an image for ROI extraction
    # (still in the original T1 orientation)
    roi_img = np.copy(imlb)

    # sum of all voxel values in any given ROI
    roisum = {}
    # sum of the mask values <0,1> used for getting the mean value of any ROI
    roimsk = {}
    for k in amyroi.rois.keys():
        roisum[k] = []
        roimsk[k] = []

    # extract the ROI values from PV corrected and uncorrected PET images
    for k in amyroi.rois.keys():
        roi_img[:] = 0
        for i in amyroi.rois[k]:
            roi_img[imlb==i] = 1.
        # now save the image mask to nii file <froi1>
        froi1 = os.path.join(prcl_dir, lbpth[1].split('.')[0][:8]+'_'+k+'.nii.gz')
        nii_roi = nib.Nifti1Image(roi_img, A)
        nii_roi.header['cal_max'] = 1.
        nii_roi.header['cal_min'] = 0.
        nib.save(nii_roi, froi1)
        # file name for the resampled ROI to PET space
        froi2 = os.path.join(prcl_dir, os.path.basename(datain['T1lbl']).split('.')[0]+ '_toPET_'+k+'.nii.gz')
        if not use_stored and os.path.isfile( Cnt['RESPATH'] ):
            cmd = [Cnt['RESPATH'],
                '-ref', fpet,
                '-flo', froi1,
                '-trans', faff,
                '-res', froi2]#,
                #'-pad', '0']
            if not Cnt['VERBOSE']: cmd.append('-voff') 
            call(cmd)
        # get the resampled ROI mask image
        rmsk = imio.getnii(froi2)
        rmsk[rmsk>1.] = 1.
        rmsk[rmsk<0.] = 0.

        # erode the cerebral white matter region if no PVC image
        if k=='wm' and not impvc:
            if Cnt['VERBOSE']: print 'i> eroding white matter as PET image not partial volume corrected.'
            nilb = nib.load(froi2)
            B = nilb.get_sform()
            tmp = ndi.filters.gaussian_filter(rmsk, imio.fwhm2sig(12,Cnt['VOXY']), mode='mirror')
            rmsk = np.float32(tmp>0.85)
            froi3 = os.path.join(prcl_dir, os.path.basename(datain['T1lbl']).split('.')[0]+ '_toPET_'+k+'_eroded.nii.gz')
            savenii(rmsk, froi3, B, Cnt)

        # ROI value and mask sums:
        rvsum  = np.sum(im*rmsk)
        rmsum  = np.sum(rmsk)
        roisum[k].append( rvsum )
        roimsk[k].append( rmsum )
        if Cnt['VERBOSE']: 
            print ''
            print '================================================================'
            print 'i> ROI extracted:', k
            print '   > value sum:', rvsum
            print '   > # voxels :', rmsum
            print '================================================================'
    # --------------------------------------------------------------

    return roisum, roimsk



def roi_extraction_spm(imdic, amyroi, datain, Cnt, dirout, r_prefix='r_', use_stored=False):

    fpet = imdic['fpet']
    M  = imdic['affine'] # matrix with affine parameters
    im   = imdic['im']

    if Cnt['VERBOSE']: print 'i> extracting ROI values'

    # split the labels path
    lbpth = os.path.split(datain['T1lbl'])
    prcl_dir = os.path.dirname(datain['T1lbl'])

    dscrp = nimpa.prc.getnii_descr(fpet)
    impvc = False
    if 'pvc' in dscrp.keys():       
        impvc = True
        if Cnt['VERBOSE']: print 'i> the provided image is partial volume corrected.'

    #------------------------------------------------------------------------------------------------------------  
    # next steps after having sorted out the upsampled input image and the rigid transformation for the T1w -> PET:

    # Get the labels before resampling to PET space so that the regions can be separated (more difficult to separate after resampling)
    if os.path.isfile(datain['T1lbl']):
        if datain['T1lbl'][-3:]=='.gz':
            flbl = imio.nii_ugzip(datain['T1lbl'])
        else:
            flbl = datain['T1lbl']
    else:
        print 'e> parcellation label image not present!'
        sys.exit()

    # ===============================================================
    # resample the labels to upsampled PET
    if fpet[-3:]=='.gz': 
        fpet_ = imio.nii_ugzip(fpet)
    else:
        fpet_ = fpet
    # get file of coregistered labels to upsampled PET
    flblu = spm_resample(fpet_, flbl, M, intrp=0, dirout=dirout, r_prefix=r_prefix, del_flo_uncmpr=True, del_out_uncmpr=True)

    imlbl = imio.getnii(flblu)

    # get the segmentation/parcellation by creating an image for ROI extraction
    roi_img = np.copy(imlbl)

    # sum of all voxel values in any given ROI
    roisum = {}
    # sum of the mask values <0,1> used for getting the mean value of any ROI
    roimsk = {}
    for k in amyroi.rois.keys():
        roisum[k] = []
        roimsk[k] = []

    # extract the ROI values from PV corrected and uncorrected PET images
    for k in amyroi.rois.keys():
        roi_img[:] = 0
        for i in amyroi.rois[k]:
            roi_img[imlbl==i] = 1.

        # ROI value and mask sums:
        rvsum  = np.sum(im*roi_img)
        rmsum  = np.sum(roi_img)
        roisum[k].append( rvsum )
        roimsk[k].append( rmsum )
        if Cnt['VERBOSE']: 
            print ''
            print '================================================================'
            print 'i> ROI extracted:', k
            print '   > value sum:', rvsum
            print '   > # voxels :', rmsum
            print '================================================================'
    # --------------------------------------------------------------

    return roisum, roimsk



# # <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
# # U T E  +   p C T    B O T S T R A P   R E C O N S T R U C T I O N   &   A N A L Y S I S
# # <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
# def btproi(datain, amyroi, txLUT, axLUT, Cnt, Bitr = 50, upscale=True, t0=0, t1=0, itr=4, fwhm=0., fcomment='', int_order=0):

#     # ff not bootstrap then do only one reconstruction
#     if Cnt['BTP']==0: Bitr = 1

#     # array for all bootstrap/normal SUVr's extracted from multiple/single reconstruction(s) and for different reference regions
#     # SUVr = { 'ute':{'nopvc':[], 'pvc':[]}, 'pct':{'nopvc':[], 'pvc':[]} }
#     # SUVr['ute']['nopvc'] = np.zeros((Bitr, len(amyroi.refi), len(amyroi.roii)), dtype=np.float32)
#     # SUVr['ute']['pvc']   = np.zeros((Bitr, len(amyroi.refi), len(amyroi.roii)), dtype=np.float32)
#     # SUVr['pct']['nopvc'] = np.zeros((Bitr, len(amyroi.refi), len(amyroi.roii)), dtype=np.float32)
#     # SUVr['pct']['pvc']   = np.zeros((Bitr, len(amyroi.refi), len(amyroi.roii)), dtype=np.float32)

#     btpout = namedtuple('btpout',  'Rpct,  Rpctpvc,  Rute,  Rutepvc, \
#                                     M1pct, M1pctpvc, M1ute, M1utepvc, \
#                                     M2pct, M2pctpvc, M2ute, M2utepvc, Itrp, Itru')
#     btpout.Rpct    = np.zeros((Bitr, len(amyroi.refi), len(amyroi.roii)), dtype=np.float32)
#     btpout.Rpctpvc = np.zeros((Bitr, len(amyroi.refi), len(amyroi.roii)), dtype=np.float32)
#     btpout.Rute    = np.zeros((Bitr, len(amyroi.refi), len(amyroi.roii)), dtype=np.float32)
#     btpout.Rutepvc = np.zeros((Bitr, len(amyroi.refi), len(amyroi.roii)), dtype=np.float32)

#     btpout.Itrp = 0
#     btpout.Itru = 0

#     for bitr in range(Bitr):
#         if Cnt['BTP']==0:
#             cmmnt = '(P)'
#         else:
#             cmmnt = '(b'+str(bitr)+')'
#         # perform reconstruction with UTE and pCT mu-maps while aligning the pCT/T1w to PET-UTE
#         recout = nipet.img.prc.osemduo(datain, txLUT, axLUT, Cnt, t0=t0, t1=t1, itr=itr, fcomment=fcomment+cmmnt, use_stored=True)
        
#         # get the PSF kernel for PVC
#         krnlPSF = nimpa.psf_measured(scale=2)
#         # trim PET and upsample.  The trimming is done based on the first file in the list below, i.e., recon with pCT
#         imudic = trimim( datain, Cnt, fpets=[recout.fpct, recout.fute], scale=2, 
#                                         fcomment='trim_'+fcomment, int_order=int_order)
#         # create separate dictionary for UTE and pCT reconstructions
#         imupct = {'im':imudic['im'][0,:,:,:], 'fpet':imudic['fpet'][0], 'affine':imudic['affine']} 
#         imuute = {'im':imudic['im'][1,:,:,:], 'fpet':imudic['fpet'][1], 'affine':imudic['affine']}

#         if bitr==0 and Cnt['BTP']>0:
#             # initialise the online variance images in this first iteration
#             btpout.M1pct   = np.zeros(imupct['im'].shape, dtype=np.float32)
#             btpout.M1pctpvc= np.zeros(imupct['im'].shape, dtype=np.float32)
#             btpout.M1ute   = np.zeros(imuute['im'].shape, dtype=np.float32)
#             btpout.M1utepvc= np.zeros(imuute['im'].shape, dtype=np.float32)
#             btpout.M2pct   = np.zeros(imupct['im'].shape, dtype=np.float32)
#             btpout.M2pctpvc= np.zeros(imupct['im'].shape, dtype=np.float32)
#             btpout.M2ute   = np.zeros(imuute['im'].shape, dtype=np.float32)
#             btpout.M2utepvc= np.zeros(imuute['im'].shape, dtype=np.float32)
        
#         #-----------------------------
#         # PCT ANALYSIS
#         if Cnt['BTP']==0:
#             cmmnt = '(pct-P)'
#         else:
#             cmmnt = '(pct-b'+str(bitr)+')'
#         # perform PVC, iterative Yang
#         pvcdic = nipet.img.prc.pvc_iYang(datain, Cnt, imupct, amyroi, krnlPSF, faffu=recout.faff, fcomment=cmmnt)
#         # the rigid transformation is the same for PVC corrected and uncorrected images
#         imupct['faff'] = pvcdic['faff']
#         # and now for PVC image
#         roisum, roimsk = nipet.img.prc.roi_extraction(pvcdic, amyroi, datain, Cnt, use_stored=False)
#         for i in range(len(amyroi.refi)):
#             refc = np.array(roisum[amyroi.refi[i]]) / np.array(roimsk[amyroi.refi[i]])
#             for j in range(len(amyroi.roii)):
#                 btpout.Rpctpvc[bitr,i,j] = ( np.array(roisum[amyroi.roii[j]]) / np.array(roimsk[amyroi.roii[j]]) ) / refc
#         # extract the ROI values from uncorrected image
#         roisum, roimsk = nipet.img.prc.roi_extraction(imupct, amyroi, datain, Cnt, use_stored=True)
#         for i in range(len(amyroi.refi)):
#             ref = np.array(roisum[amyroi.refi[i]]) / np.array(roimsk[amyroi.refi[i]])
#             for j in range(len(amyroi.roii)):
#                 btpout.Rpct[bitr,i,j] = ( np.array(roisum[amyroi.roii[j]]) / np.array(roimsk[amyroi.roii[j]]) ) / ref
#         # calculate variance online
#         if Cnt['BTP']>0 and imupct['im'].shape==btpout.M1pct.shape:
#             nipet.mmr_auxe.varon(btpout.M1pct,    btpout.M2pct,    imupct['im']/ref , btpout.Itrp, Cnt)
#             nipet.mmr_auxe.varon(btpout.M1pctpvc, btpout.M2pctpvc, pvcdic['im']/refc, btpout.Itrp, Cnt)
#             btpout.Itrp += 1
#         else:
#             print 'e> Omitting on-line variance calculations. Check the shape of images.'


#         #-----------------------------
#         # UTE ANALYSIS
#         if Cnt['BTP']==0:
#             cmmnt = '(ute-P)'
#         else:
#             cmmnt = '(ute-b'+str(bitr)+')' 
#         # perform PVC, iterative Yang
#         pvcdic = nipet.img.prc.pvc_iYang(datain, Cnt, imuute, amyroi, krnlPSF, faffu=recout.faff, fcomment=cmmnt)
#         # the rigid transformation is the same for PVC corrected and uncorrected images
#         imuute['faff'] = pvcdic['faff']
#         # extract the ROI values from uncorrected image
#         roisum, roimsk = nipet.img.prc.roi_extraction(imuute, amyroi, datain, Cnt, use_stored=False)
#         for i in range(len(amyroi.refi)):
#             ref = np.array(roisum[amyroi.refi[i]]) / np.array(roimsk[amyroi.refi[i]])
#             for j in range(len(amyroi.roii)):
#                 btpout.Rute[bitr,i,j] = ( np.array(roisum[amyroi.roii[j]]) / np.array(roimsk[amyroi.roii[j]]) ) / ref
#         # and now for PVC image
#         roisum, roimsk = nipet.img.prc.roi_extraction(pvcdic, amyroi, datain, Cnt, use_stored=True)
#         for i in range(len(amyroi.refi)):
#             ref = np.array(roisum[amyroi.refi[i]]) / np.array(roimsk[amyroi.refi[i]])
#             for j in range(len(amyroi.roii)):
#                 btpout.Rutepvc[bitr,i,j] = ( np.array(roisum[amyroi.roii[j]]) / np.array(roimsk[amyroi.roii[j]]) ) / ref
#         # calculate variance online
#         if Cnt['BTP']>0 and imuute['im'].shape==btpout.M1ute.shape:
#             nipet.mmr_auxe.varon(btpout.M1ute,    btpout.M2ute,    imuute['im'], btpout.Itru, Cnt)
#             nipet.mmr_auxe.varon(btpout.M1utepvc, btpout.M2utepvc, pvcdic['im'], btpout.Itru, Cnt)
#             btpout.Itru += 1
#         else:
#             print 'e> different shape encountered in online variance calculations (UTE).  Omitting.'

#     return btpout


#!/usr/bin/env python
"""initialise the NIMPA package (part of NiftyPET package)"""
__author__      = "Pawel Markiewicz"
__copyright__   = "Copyright 2018 Pawel Markiewicz @ University College London"
#------------------------------------------------------------------------------

import os
import sys
import platform

# if using conda put the resources in the folder with the environment name
if 'CONDA_DEFAULT_ENV' in os.environ:
	env = os.environ['CONDA_DEFAULT_ENV']
	print 'i> conda environment found:', env
else:
	env = ''
# create the path for the resources files according to the OS platform
if platform.system() in ['Linux', 'Darwin']:
	path_resources = os.path.join( os.path.join(os.path.expanduser('~'),   '.niftypet'), env )
elif platform.system() == 'Windows' :
	path_resources = os.path.join( os.path.join(os.getenv('LOCALAPPDATA'), '.niftypet'), env )
else:
	print 'e> unrecognised operating system!'
	
sys.path.append(path_resources)
try:
    import resources
except ImportError as ie:
    print '----------------------------'
    print 'e> Import Error: NiftyPET''s resources file <resources.py> could not be imported.  It should be in ''~/.niftypet/resources.py'' (Linux) or ''//Users//USERNAME//AppData//Local//niftypet//resources.py'' (Windows) but likely it does not exists.'
    print '----------------------------'
#===========================

if resources.CC_ARCH != '' and platform.system() in ['Linux', 'Windows']:
	from dinf import gpuinfo, dev_info
	
from prc import trimim, iyang, pvc_iyang, psf_general, psf_measured

from prc import realign_mltp_spm, resample_mltp_spm
from prc import coreg_spm, coreg_vinci, resample_spm, resample_vinci
from prc import affine_fsl, resample_fsl
from prc import affine_niftyreg, resample_niftyreg, pet2pet_rigid
from prc import create_dir, time_stamp, fwhm2sig, getnii, getnii_descr, array2nii, dcm2im
from prc import orientnii, nii_ugzip, nii_gzip, niisort, dcmsort, dcminfo, dcmanonym

from prc import dice_coeff, dice_coeff_multiclass
from prc import imfill, create_mask
from prc import bias_field_correction
from prc import pick_t1w

from prc import motion_reg

from prc import ct2mu
from prc import nii_modify

from prc import dcm2nii

from img import create_disk, profile_points



"""Resources file for NiftyPET NIPET and NIMPA etc."""
#---------------------------------------------------------------
__author__      = "Pawel Markiewicz"
__copyright__   = "Copyright 2018"
#---------------------------------------------------------------

import numpy as np
from math import pi
import platform
import os

# Hardware (CT-based) mu-maps, which come with the mMR scanner.
# The names may be different
hrdwr_mu = [   
    'umap_HNMCL_10606489.v.hdr',   # (1) Head and neck lower coil
    'umap_HNMCU_10606489.v.hdr',   # (2) Head and neck upper coil
    'umap_SPMC_10606491.v.hdr',    # (3) Spine coil
    'umap_PT_2291734.v.hdr'        # (4) Table
]

# Radioisotope look-up table
riLUT = {   'Ge68':{'BF':0.891,'thalf':270.9516*24*60*60},
            'Ga68':{'BF':0.891,'thalf':67.71*60},
            'F18':{'BF':0.967, 'thalf':109.77120*60},
            'C11':{'BF':0.998, 'thalf':20.38*60},
            'O15':{'BF':0.999, 'thalf':122.2416}
        }

# -----------------------------------------------------
# The name and path to the NiftyPET tools (software)
DIRTOOLS  = 'NiftyPET_tools'
MSVC_VRSN = 'Visual Studio 12 2013 Win64'
CMAKE_TLS_PAR = '' #-DUSE_SSE=OFF'
# PATHTOOLS = os.path.join('/chosen/path/', DIRTOOLS)
#> path to Python wrapper of Vinci
VINCIPATH = ''
# -----------------------------------------------------

# -----------------------------------------------------
# DO NOT MODIFY BELOW--DONE AUTOMATICALLY
### start GPU properties ###
### end GPU properties ###

# paths to apps and tools needed by NiftyPET
### start NiftyPET tools ###
### end NiftyPET tools ###
# -----------------------------------------------------

# enable xnat module
ENBLXNAT = False
# enable Agg
ENBLAGG = False
# compile DCM2NIIX, otherwise download a compiled version for the system used
CMPL_DCM2NIIX = False

#============ SIEMENS mMR SCANNER C O N S T A N T S ===============
# number of rings (axially) and crystals (transaxially)
NRNG = 64

# number of crystals transaxially
NCRS = 504

# reduced number of crystals by the gaps (dead crystals)
NCRSR = 448

# maximum ring difference
MRD = 60

# number of linear indexes for 1/2 Michelogram (NRNG**2/2 + NRNG/2)
NLI2R = 2080-6

#number of angular indexes in a 2D sinogram
A = 252
#number of bin indexes in a 2D sino
W = 344
H = W/2

NSN11 = 837
NSN1  = 4084
NSN64 = NRNG*NRNG
# 0: SSRB, 1: span-1, or 11: span-11 (11, default)
SPAN = 11

RNG_STRT = 0
RNG_END = 64

# no of sinos in a segment out of 11 segments
seg   = np.array([127,115,115,93,93,71,71,49,49,27,27])

# minimum and maximum ring difference for each segment
minrd = np.array([-5,-16, 6,-27,17,-38,28,-49,39,-60,50])
maxrd = np.array([ 5, -6,16,-17,27,-28,38,-39,49,-50,60])
#----------

#------------------------------------------------------
# scatter ring definition
sct_irng = np.int16([0, 10, 19, 28, 35, 44, 53, 63])
# number of scatter rings (used for scatter estimation)
NSRNG = len(sct_irng)
#------------------------------------------------------

# number of direct sinograms (i.e., for segment 0)
SEG0 = 127

# Reference image size (usually the default from Siemens) and GPU dimensions for optimal execution
#~~~
SO_IMZ = 127
SO_IMY = 344
SO_IMX = 344
SO_VXX = 0.208626
SO_VXZ = 0.203125
SO_VXY = SO_VXX

SZ_IMZ = 127
SZ_IMY = 320
SZ_IMX = 320
SZ_VOXY = 0.208626
SZ_VOXZ = 0.203125
#~~~
# SO_IMZ = 127
# SO_IMY = 384
# SO_IMX = 384
# SO_VXX = 0.1669
# SO_VXZ = 0.203125
# SO_VXY = SO_VXX

# SZ_IMZ = 127
# SZ_IMY = 384
# SZ_IMX = 384
# SZ_VOXY = 0.1669
# SZ_VOXZ = 0.203125
#~~~

#~~~
# inverse size
SZ_VOXZi = round(1/SZ_VOXZ,6)
# squared radius of the transaxial field of view
TFOV2 = 890.0

#-------Scatter image size in x,y,z directions
# target scale factors for scatter mu-map and emission image respectively
# transmission (mu-map)
TRGTSCT = [0.5, 0.33]
SS_IMX = int(np.ceil(TRGTSCT[0]*SO_IMX)//2*2)
SS_IMY = int(np.ceil(TRGTSCT[0]*SO_IMY)//2*2)
SS_IMZ = int(np.ceil(TRGTSCT[0]*SO_IMZ)//2*2-1)
SS_VXY = round((SO_VXY*SO_IMX)/SS_IMX,6)
SS_VXZ = round((SO_VXZ*SO_IMZ)/SS_IMZ,6)
IS_VXZ = round(1/SS_VXZ,6)
# scaling [z,y,x]
SCTSCLMU = [float(SS_IMZ)/SO_IMZ, float(SS_IMY)/SO_IMY, float(SS_IMX)/SO_IMX]
# emission
SSE_IMX = int(np.ceil(TRGTSCT[1]*SO_IMX)//2*2)
SSE_IMY = int(np.ceil(TRGTSCT[1]*SO_IMY)//2*2)
SSE_IMZ = int(np.ceil(TRGTSCT[1]*SO_IMZ)//2*2+1)
SSE_VXY = round((SO_VXY*SO_IMX)/SSE_IMX,6)
SSE_VXZ = round((SO_VXZ*SO_IMZ)/SSE_IMZ,6)
# scaling [z,y,x]
SCTSCLEM = [float(SSE_IMZ)/SO_IMZ, float(SSE_IMY)/SO_IMY, float(SSE_IMX)/SO_IMX]


# # scaling for the emission image [z,y,x]
# SCTSCLEM = [0.34, 0.33, 0.33]
# # scaling for the mu-map
# SCTSCLMU = [0.499, 0.5, 0.5]

# SS_IMX = int(np.ceil(SCTSCLMU[2]*SO_IMX)//2*2)#172
# SS_IMY = int(np.ceil(SCTSCLMU[1]*SO_IMY)//2*2)#172
# SS_IMZ = int(np.ceil(SCTSCLMU[0]*SO_IMZ)//2*2-1)#63
# SS_VXY = round((SO_VXY*SO_IMX)/SS_IMX,6) # 0.417252 #
# SS_VXZ = round((SO_VXZ*SO_IMZ)/SS_IMZ,6) # 0.409474 #
# IS_VXZ = round(1/SS_VXZ,6)

# SSE_IMX = int(np.ceil(SCTSCLEM[2]*SO_IMX)//2*2) #114
# SSE_IMY = int(np.ceil(SCTSCLEM[1]*SO_IMY)//2*2) #114
# SSE_IMZ = int(np.ceil(SCTSCLEM[0]*SO_IMZ)//2*2-1) #43

# SSE_VXY = round((SO_VXY*SO_IMX)/SSE_IMX,6) #0.629538
# SSE_VXZ = round((SO_VXZ*SO_IMZ)/SSE_IMZ,6) #0.599927
#-------

#> decay correction
DCYCRR = True

#--- Time of Flight ---
#speed of light
CLGHT = 29979245800 #cm/s
#coincidence time window [ps]
CWND = 5859.38e-12

#number of TOF bins
TOFBINN = 1
#size of TOF bin in [ps]
TOFBINS = 390e-12
#size of TOF BIN in cm of travelled distance
TOFBIND = TOFBINS*CLGHT
#inverse of the above
ITOFBIND = 1/TOFBIND


#ring radius
R = 32.8
#effective ring radius accounting for the depth of interaction
RE = (R + 0.67)#0.67

#axial crystal width
AXR = 0.40625

#crystal angle
ALPHA = 0.714286*pi/180 #2*pi/NCRS

#crystal gap period
TGAP = 9

#crystal gap offset (used for getting the sino gaps right at the position)
OFFGAP = 1

#--- FOR SCATTER ---
#electron radius **2
R02 = 7.940787449825884e-26
#detection lower energy threashold
LLD = 430000
E511 = 511008
#energy resolution
ER = 0#0.154
#discretisation of the scatter angle spectrum
NCOS = 256
#cosine of maximum allowed scatter angle
COSUPSMX = 0.725 #0.58 #0.722 #Elow = E511/(2-cos(upsmx))
#step of the discreatisation
COSSTP = (1-COSUPSMX)/(NCOS-1)
#inverse of the step
ICOSSTP = 1/COSSTP

# intensity percentage threshold of voxels to be considered in the image
ETHRLD = 0.05

#=================================================================================================
def get_gpu_constants(Cnt = {}):
    '''Return a dictionary of GPU related constants'''
    if 'DEV_ID'  in globals():  Cnt['DEVID']  = DEV_ID      # device id; used for choosing the GPU device for calculations
    if 'CC_ARCH' in globals():  Cnt['CCARCH'] = CC_ARCH     # chosen device architectures for NVCC compilation
    
    return Cnt

#=================================================================================================
def get_setup(Cnt = {}):
    '''Return a dictionary of GPU, mu-map hardware and third party set-up.'''
    
    # the name of the folder for NiftyPET tools
    Cnt['DIRTOOLS'] = DIRTOOLS

    # additional paramteres for compiling tools with cmake
    Cnt['CMAKE_TLS_PAR'] = CMAKE_TLS_PAR

    # hardware mu-maps
    Cnt['HMULIST'] = hrdwr_mu
    
    # Microsoft Visual Studio Compiler version
    Cnt['MSVC_VRSN'] = MSVC_VRSN

    # GPU related setup
    Cnt = get_gpu_constants(Cnt)
    
    if 'PATHTOOLS'  in globals() and PATHTOOLS!='':     Cnt['PATHTOOLS'] = PATHTOOLS
    # image processing setup
    if 'RESPATH'    in globals() and RESPATH!='':       Cnt['RESPATH']  = RESPATH
    if 'REGPATH'    in globals() and REGPATH!='':       Cnt['REGPATH']  = REGPATH
    if 'DCM2NIIX'   in globals() and DCM2NIIX!='':      Cnt['DCM2NIIX'] = DCM2NIIX
    # hardware mu-maps
    if 'HMUDIR'     in globals() and HMUDIR!='':        Cnt['HMUDIR']   = HMUDIR
    if 'VINCIPATH'  in globals() and VINCIPATH!='':     Cnt['VINCIPATH'] = VINCIPATH

    Cnt['ENBLXNAT'] = ENBLXNAT
    Cnt['ENBLAGG'] = ENBLAGG
    Cnt['CMPL_DCM2NIIX'] = CMPL_DCM2NIIX

    return Cnt

#=================================================================================================
def get_mmr_constants():
    '''
    Put all the constants together in a dictionary
    '''

    Cnt = {
        'ISOTOPE':'F18',
        'DCYCRR':DCYCRR,
        'ALPHA':ALPHA,
        'NRNG':NRNG,
        'NSRNG':NSRNG,
        'NCRS':NCRS,
        'NCRSR':NCRSR,
        'NBCKT':224,
        'NSANGLES':A,
        'NSBINS':W,
        'Naw':-1,   # number of total active bins per 2D sino
        'NSN11': NSN11, # number of sinos in span-11
        'NSN1': NSN1, # number of sinos in span-1
        'NSN64': NSN64, # number of sinos in span-1 with no MRD limit
        'MRD': MRD, # maximum ring difference RD
        'SPN':SPAN, # span-1 (1), span-11 (11), ssrb (0)
        'TFOV2':TFOV2, # squared radius of TFOV

        'RNG_STRT':RNG_STRT, # limit axial extension by defining start and end ring
        'RNG_END' :RNG_END,  # this feature only works with span-1 processing (Cnt['SPN']=1)

        'SS_IMZ':SS_IMZ, #Scatter mu-map iamge size 
        'SS_IMY':SS_IMY,
        'SS_IMX':SS_IMX,
        'SS_VXZ':SS_VXZ,
        'SS_VXY':SS_VXY,
        'IS_VXZ':IS_VXZ,
        'SSE_IMZ':SSE_IMZ, #Scatter emission image size
        'SSE_IMY':SSE_IMY,
        'SSE_IMX':SSE_IMX,
        'SSE_VXZ':SSE_VXZ,
        'SSE_VXY':SSE_VXY,
        'SZ_IMZ':SZ_IMZ, #GPU optimised image size
        'SZ_IMY':SZ_IMY,
        'SZ_IMX':SZ_IMX,
        'SZ_VOXZ':SZ_VOXZ,
        'SZ_VOXY':SZ_VOXY,
        'SZ_VOXZi':SZ_VOXZi,
        'SO_IMZ':SO_IMZ, #Original image size (from Siemens)
        'SO_IMY':SO_IMY,
        'SO_IMX':SO_IMX,
        'SO_VXZ':SO_VXZ,
        'SO_VXY':SO_VXY,
        'SO_VXX':SO_VXX,
        'NSEG0':SEG0,
        'RE':RE, #effective ring radius
        'R':R,
        'SEG':seg,
        'MNRD':minrd,
        'MXRD':maxrd,
        'SCTRNG':sct_irng,
        'TGAP':TGAP,
        'OFFGAP':OFFGAP,
        'AXR':AXR,
        'R02':R02, #squared electron radius
        'LLD':LLD, #lower energy threashold
        'E511':E511,
        'ER':ER, #energy resolution
        'COSUPSMX':COSUPSMX, #cosine of max allowed scatter angle
        'NCOS':NCOS, #number of cos samples for LUT
        'COSSTP':COSSTP, #cosine step
        'ICOSSTP':ICOSSTP, #inverse of cosine step
        'ETHRLD':ETHRLD, #intensity emission image threshold (used in scatter modelling)
        'CLGHT':CLGHT, #speed of light [cm/s]
        'CWND':CWND,    #coincidence time window [ps]
        'TOFBINN':TOFBINN, #number of TOF bins
        'TOFBINS':TOFBINS,   #TOF bin width [ps]
        'TOFBIND':TOFBIND,
        'ITOFBIND':ITOFBIND,
        # affine and image size for the reconstructed image, assuming the centre of voxels in mm
        'AFFINE':np.array([ [-10*SO_VXX, 0., 0., 5.*SO_IMX*SO_VXX ], #+5.*SO_VXX
                            [0., 10*SO_VXY, 0., -5.*SO_IMY*SO_VXY ], #+5.*SO_VXY
                            [0., 0., 10*SO_VXZ, -5.*SO_IMZ*SO_VXZ ], #-5.*SO_VXZ
                            [0., 0., 0., 1.]]),
        'IMSIZE':np.array([SO_IMZ, SO_IMY, SO_IMX]),


        'BTP':0, #1:non parametric bootstrap, 2: parametric bootstrap (recommended)
        'BTPRT':1.0, # Ratio of bootstrapped/original events (enables downsampling)

        'VERBOSE':False,

        'SCTSCLEM':SCTSCLEM,
        'SCTSCLMU':SCTSCLMU,
    }

    # get the setup for GPU and third party apps
    Cnt = get_setup(Cnt=Cnt)

    return Cnt



# initialise the module folder

from prc import trimim, iyang, pvc_iyang, psf_general, psf_measured
from prc import ct2mu
from prc import nii_modify
from prc import bias_field_correction

from imio import create_dir, time_stamp, fwhm2sig, getnii, getnii_descr, array2nii
from imio import orientnii, nii_ugzip, nii_gzip, dcmsort, niisort, dcm2im, dcminfo
from imio import dcmanonym, pick_t1w, dcm2nii

from regseg import realign_mltp_spm, resample_mltp_spm

from regseg import coreg_spm, resample_spm, affine_niftyreg, resample_niftyreg
from regseg import coreg_vinci, resample_vinci
from regseg import motion_reg
from regseg import dice_coeff, dice_coeff_multiclass
from regseg import imfill, create_mask

from regseg import affine_fsl, resample_fsl

# will be depreciated
from prc import pet2pet_rigid



""" NIMPA: functions for neuro image processing and analysis
    Generates images.
"""
__author__    = "Pawel Markiewicz"
__copyright__ = "Copyright 2019"
#-------------------------------------------------------------------------------

import sys
import os

import numpy as np
import math
import scipy.ndimage as ndi




#=================================================
def create_disk(shape_in, r=1, a=0, b=0, gen_scale=1, threshold=None):

    if len(shape_in)==2:
        shape = (1,)+shape_in
    if len(shape_in)==3:
        shape = shape_in
    
    imsk = np.zeros((gen_scale*shape[1], gen_scale*shape[2]), dtype=np.float32)
    for t in np.arange(0, math.pi, math.pi/(gen_scale*400)):
        x = gen_scale*r*np.cos(t) + gen_scale*a
        y = gen_scale*r*np.sin(t) + gen_scale*b
        
        for ys in np.arange(-y+2*b*gen_scale, y, 0.5):
            v = 0.5*gen_scale*shape[1] - np.ceil(ys)
            u = 0.5*gen_scale*shape[2] + np.floor(x)
            imsk[v.astype(np.int16), u.astype(np.int16)] = 1.

    if gen_scale>1:
        imsk = ndi.interpolation.zoom(imsk,gen_scale**-1,order=1)

    if threshold:
        imsk = imsk>threshold
    
    if len(shape_in)==3:
        msk = np.repeat(imsk.reshape((1, shape[1], shape[1])), shape[0], axis=0)
    elif len(shape_in)==2:
        msk = imsk
    return msk
#=================================================



#=================================================
def profile_points(im, p0, p1, steps=100):

    p = np.array([p1[0]-p0[0], p1[1]-p0[1]])
    nrm = np.sum(p**2)**.5
    p = p/nrm

    tt = np.linspace(0, nrm, steps)
    profile = np.zeros(len(tt), dtype=im.dtype)

    c = 0
    for t in tt:
        u = t*p[0] + p0[0]
        v = t*p[1] + p0[1]

        profile[c] = im[int(round(v)), int(round(u))]
        c+=1

    return profile
#=================================================


""" NIMPA: functions for neuro image processing and analysis.
	Includes functions relating to image registration/segmentation.
    
"""
__author__    = "Pawel Markiewicz"
__copyright__ = "Copyright 2019"
#-------------------------------------------------------------------------------

import sys
import os
import shutil
from subprocess import call
import glob

import numpy as np
import scipy.ndimage as ndi

import imio
import prc



def imfill(immsk):
    '''fill the empty patches of image mask 'immsk' '''

    for iz in range(immsk.shape[0]):
        for iy in range(immsk.shape[1]):
            ix0 = np.argmax(immsk[iz,iy,:]>0)
            ix1 = immsk.shape[2] - np.argmax(immsk[iz,iy,::-1]>0)
            if (ix1-ix0) > immsk.shape[2]-10: continue
            immsk[iz,iy,ix0:ix1] = 1
    return immsk



#-------------------------------------------------------------------------------
# Create object mask for the input image 
#-------------------------------------------------------------------------------
def create_mask(
        fnii,
        fimout = '',
        outpath = '',
        fill = 1,
        dtype_fill = np.uint8,
        
        thrsh = 0.,
        fwhm = 0.,):

    ''' create mask over the whole image or over the threshold area'''
    

    #> output path
    if outpath=='' and fimout!='':
        opth = os.path.dirname(fimout)
        if opth=='':
            opth = os.path.dirname(fnii)
            fimout = os.path.join(opth, fimout)

    elif outpath=='':
        opth = os.path.dirname(fnii)

    else:
        opth = outpath

    #> output file name if not given
    if fimout=='':
        fniis = os.path.split(fnii)
        fimout = os.path.join(opth, fniis[1].split('.nii')[0]+'_mask.nii.gz')

    niidct = imio.getnii(fnii, output='all')
    im = niidct['im']
    hdr = niidct['hdr']

    if im.ndim>3:
        raise ValueError('The masking function only accepts 3-D images.')

    #> generate output image
    if thrsh>0.:
        smoim = ndi.filters.gaussian_filter(
                    im,
                    imio.fwhm2sig(fwhm, voxsize=abs(hdr['pixdim'][1])), 
                    mode='mirror')
        thrsh = thrsh*smoim.max()
        immsk = np.int8(smoim>thrsh)
        immsk = imfill(immsk)

        #> output image
        imo = fill * immsk.astype(dtype_fill)

    else:

        imo = fill * np.ones(im.shape, dtype = dtype_fill)

    #> save output image
    imio.array2nii( 
                imo,
                niidct['affine'],
                fimout,
                trnsp = (niidct['transpose'].index(0),
                         niidct['transpose'].index(1),
                         niidct['transpose'].index(2)),
                flip = niidct['flip'])

    return {'fim':fimout, 'im':imo}
#-------------------------------------------------------------------------------



# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
# I M A G E   R E G I S T R A T I O N
# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>


#-------------------------------------------------------------------------------
def affine_niftyreg(
    fref,
    fflo,
    outpath='',
    fname_aff='',
    pickname='ref',
    fcomment='',
    executable = '',
    omp=1,
    rigOnly = False,
    affDirect = False,
    maxit=5,
    speed=True,
    pi=50, pv=50,
    smof=0, smor=0,
    rmsk=True,
    fmsk=True,
    rfwhm=15.,
    rthrsh=0.05,
    ffwhm = 15.,
    fthrsh=0.05,
    verbose=True):

    # check if the executable exists:
    if not os.path.isfile(executable):
        raise IOError('Incorrect path to executable file for registration.')

    #create a folder for images registered to ref
    if outpath!='':
        odir = os.path.join(outpath,'affine-niftyreg')
        fimdir = os.path.join(outpath, os.path.join('affine-niftyreg','mask'))
    else:
        odir = os.path.join(os.path.dirname(fflo),'affine-niftyreg')
        fimdir = os.path.join(os.path.dirname(fflo), 'affine-niftyreg', 'mask')
    imio.create_dir(odir)
    imio.create_dir(fimdir)

    if rmsk:
        f_rmsk = os.path.join(fimdir, 'rmask_'+os.path.basename(fref).split('.nii')[0]+'.nii.gz')
        create_mask(fref, fimout = f_rmsk, thrsh = rthrsh, fwhm = rfwhm)
    
    if fmsk:
        f_fmsk = os.path.join(fimdir, 'fmask_'+os.path.basename(fflo).split('.nii')[0]+'.nii.gz')
        create_mask(fflo, fimout = f_fmsk, thrsh = fthrsh, fwhm = ffwhm)

    # output in register with ref and text file for the affine transform
    if fname_aff!='':
        fout = os.path.join(odir, fname_aff.split('.')[0]+'.nii.gz')
        faff = os.path.join(odir, fname_aff)
    else:
        if pickname=='ref':
            fout = os.path.join(odir, 'affine_ref-' \
                +os.path.basename(fref).split('.nii')[0]+fcomment+'.nii.gz')
            faff = os.path.join(odir, 'affine_ref-' \
                +os.path.basename(fref).split('.nii')[0]+fcomment+'.txt')
        elif pickname=='flo':
            fout = os.path.join(odir, 'affine_flo-' \
                +os.path.basename(fflo).split('.nii')[0]+fcomment+'.nii.gz')
            faff = os.path.join(odir, 'affine_flo-' \
                +os.path.basename(fflo).split('.nii')[0]+fcomment+'.txt')
    
    # call the registration routine
    cmd = [executable,
         '-ref', fref,
         '-flo', fflo,
         '-aff', faff,
         '-pi', str(pi),
         '-pv', str(pv),
         '-smooF', str(smof),
         '-smooR', str(smor),
         '-maxit', '10',
         '-omp', str(omp),
         '-res', fout]
    if speed:
        cmd.append('-speeeeed')
    if rigOnly:
        cmd.append('-rigOnly')
    if affDirect:
        cmd.append('affDirect')
    if rmsk: 
        cmd.append('-rmask')
        cmd.append(f_rmsk)
    if fmsk:
        cmd.append('-fmask')
        cmd.append(f_fmsk)
    if not verbose:
        cmd.append('-voff')

    call(cmd)

    #> affine to Numpy array
    aff = np.loadtxt(faff)
       
    return {'affine':aff, 'faff':faff, 'fim':fout} #faff, fout
#-------------------------------------------------------------------------------------

#-------------------------------------------------------------------------------------
def resample_niftyreg(
        fref,
        fflo,
        faff,
        outpath = '',
        fimout = '',
        fcomment = '',
        pickname = 'ref',
        intrp = 1,
        executable = '',
        verbose = True):

    # check if the executable exists:
    # if executable=='' and 'RESPATH' in Cnt and os.path.isfile(Cnt['RESPATH']):
    #     executable = Cnt['RESPATH']
    
    if not os.path.isfile(executable):
        raise IOError('Incorrect path to executable file for registration.')

    #> output path
    if outpath=='' and fimout!='':
        opth = os.path.dirname(fimout)
        if opth=='':
            opth = os.path.dirname(fflo)

    elif outpath=='':
        opth = os.path.dirname(fflo)

    else:
        opth = outpath

    imio.create_dir(opth)

    #> the output naming
    if '/' in fimout:
        fout = fimout
    elif fimout!='' and not os.path.isfile(fimout):
        fout = os.path.join(opth, fimout)
    elif pickname=='ref':
        fout = os.path.join(opth, 'affine_ref-' \
                + os.path.basename(fref).split('.nii')[0]+fcomment+'.nii.gz')
    elif pickname=='flo':        
        fout = os.path.join(opth, 'affine_flo-' \
                + os.path.basename(fflo).split('.nii')[0]+fcomment+'.nii.gz')

    if isinstance(intrp, (int, long)): intrp = str(intrp)
    
    cmd = [executable,
       '-ref', fref,
       '-flo', fflo,
       '-trans', faff,
       '-res', fout,
       '-inter', intrp
       ]
    if not verbose:
        cmd.append('-voff')
    call(cmd)

    return fout
#-------------------------------------------------------------------------------------



#===============================================================================
# S P M registration
#===============================================================================
def coreg_spm(
        imref,
        imflo,
        matlab_eng='',
        outpath='',
        fname_aff='',
        fcomment = '',
        pickname='ref',
        costfun='nmi',
        sep = [4,2],
        tol = [ 0.0200,0.0200,0.0200,0.0010,0.0010,0.0010,
                0.0100,0.0100,0.0100,0.0010,0.0010,0.0010],
        fwhm = [7,7],
        params = [0,0,0,0,0,0],
        graphics = 1,
        visual = 0,
        del_uncmpr=True,
        save_arr = True,
        save_txt = True,
    ):

    import matlab
    from pkg_resources import resource_filename

    #-start Matlab engine if not given
    if matlab_eng=='':
        import matlab.engine
        eng = matlab.engine.start_matlab()
    else:
        eng = matlab_eng

    # add path to SPM matlab file
    spmpth = resource_filename(__name__, 'spm')
    print 'PATH: ' + spmpth
    eng.addpath(spmpth, nargout=0)

    #> output path
    if outpath=='' and fname_aff!='' and '/' in fname_aff:
        opth = os.path.dirname(fname_aff)
        if opth=='':
            opth = os.path.dirname(imflo)
        fname_aff = os.path.basename(fname_aff)
    elif outpath=='':
        opth = os.path.dirname(imflo)
    else:
        opth = outpath
    imio.create_dir(opth)

    #> decompress ref image as necessary 
    if imref[-3:]=='.gz':
        imrefu = imio.nii_ugzip(imref, outpath=opth)
    else:
        fnm = os.path.basename(imref).split('.nii')[0] + '_copy.nii'
        imrefu = os.path.join(opth, fnm)
        shutil.copyfile(imref, imrefu)
    
    #> floating
    if imflo[-3:]=='.gz': 
        imflou = imio.nii_ugzip(imflo, outpath=opth)
    else:
        fnm = os.path.basename(imflo).split('.nii')[0] + '_copy.nii'
        imflou = os.path.join(opth, fnm)
        shutil.copyfile(imflo, imflou)

    # run the matlab SPM coregistration
    Mm, xm = eng.coreg_spm_m(
        imrefu,
        imflou,
        costfun,
        matlab.double(sep),
        matlab.double(tol),
        matlab.double(fwhm),
        matlab.double(params),
        graphics,
        visual,
        nargout=2
    )

    # get the affine matrix
    M = np.array(Mm._data.tolist())
    M = M.reshape(4,4).T

    # get the translation and rotation parameters in a vector
    x = np.array(xm._data.tolist())

    # delete the uncompressed files
    if del_uncmpr:
        if imref[-3:]=='.gz': os.remove(imrefu)
        if imflo[-3:]=='.gz': os.remove(imflou)

    imio.create_dir( os.path.join(opth, 'affine-spm') )

    #---------------------------------------------------------------------------
    if fname_aff == '':
        if pickname=='ref':
            faff = os.path.join(
                    opth,
                    'affine-spm',
                    'affine-ref-'+os.path.basename(imref).split('.nii')[0]+fcomment+'.npy')
        else:
            faff = os.path.join(
                    opth,
                    'affine-spm',
                    'affine-flo-'+os.path.basename(imflo).split('.nii')[0]+fcomment+'.npy')
    
    else:

        #> add '.npy' extension if not in the affine output file name
        if not fname_aff.endswith('.npy'):
            fname_aff += '.npy'

        faff = os.path.join(
                opth,
                'affine-spm',
                fname_aff)
    #---------------------------------------------------------------------------

    #> safe the affine transformation
    if save_arr:
        np.save(faff, M)
    if save_txt:
        np.savetxt(faff.split('.npy')[0]+'.txt', M)
    
    return {'affine':M, 'faff':faff,
            'rotations':x[3:], 'translations':x[:3],
            'matlab_eng':eng}
#===============================================================================





#===============================================================================
# S P M resampling
#===============================================================================
def resample_spm(
        imref,
        imflo,
        M,
        matlab_eng='',
        intrp=1.,
        which=1,
        mask=0,
        mean=0,
        outpath='',
        fimout='',
        fcomment='',
        prefix='r_',
        pickname='ref',
        del_ref_uncmpr=False,
        del_flo_uncmpr=False,
        del_out_uncmpr=False
    ):


    print '====================================================================='
    print ' S P M  inputs:'
    print '> ref:', imref
    print '> flo:', imflo
    print '====================================================================='

    import matlab
    from pkg_resources import resource_filename

    #-start Matlab engine if not given
    if matlab_eng=='':
        import matlab.engine
        eng = matlab.engine.start_matlab()
    else:
        eng = matlab_eng

    # add path to SPM matlab file
    spmpth = resource_filename(__name__, 'spm')
    eng.addpath(spmpth, nargout=0)

    #> output path
    if outpath=='' and fimout!='':
        opth = os.path.dirname(fimout)
        if opth=='':
            opth = os.path.dirname(imflo)

    elif outpath=='':
        opth = os.path.dirname(imflo)

    else:
        opth = outpath

    imio.create_dir(opth)

    #> decompress if necessary 
    if imref[-3:]=='.gz':
        imrefu = imio.nii_ugzip(imref, outpath=opth)
    else:
        fnm = os.path.basename(imref).split('.nii')[0] + '_copy.nii'
        imrefu = os.path.join(opth, fnm)
        shutil.copyfile(imref, imrefu)

    #> floating
    if imflo[-3:]=='.gz': 
        imflou = imio.nii_ugzip(imflo, outpath=opth)
    else:
        fnm = os.path.basename(imflo).split('.nii')[0] + '_copy.nii'
        imflou = os.path.join(opth, fnm)
        shutil.copyfile(imflo, imflou)

    if isinstance(M, basestring):
        if os.path.basename(M).endswith('.txt'):
            M = np.loadtxt(M)
            print 'i> matrix M given in the form of text file'
        elif os.path.basename(M).endswith('.npy'):
            M = np.load(M)
            print 'i> matrix M given in the form of NumPy file'
        else:
            raise IOError('e> unrecognised file extension for the affine.')
            
    elif isinstance(M, (np.ndarray, np.generic)):
        print 'i> matrix M given in the form of Numpy array'
    else:
        raise IOError('The form of affine matrix not recognised.')

    # run the Matlab SPM resampling
    r = eng.resample_spm_m(
        imrefu,
        imflou,
        matlab.double(M.tolist()),
        mask,
        mean,
        intrp,
        which,
        prefix)

    
    #-compress the output
    split = os.path.split(imflou)
    fim = os.path.join(split[0], prefix+split[1])
    imio.nii_gzip(fim, outpath=opth)

    # delete the uncompressed
    if del_ref_uncmpr:  os.remove(imrefu)
    if del_flo_uncmpr and os.path.isfile(imflou):  os.remove(imflou)
    if del_out_uncmpr: os.remove(fim)

    #> the compressed output naming
    if fimout!='':
        fout = os.path.join(opth, fimout)
    elif pickname=='ref':
        fout = os.path.join(opth, 'affine_ref-' \
                + os.path.basename(imrefu).split('.nii')[0]+fcomment+'.nii.gz')
    elif pickname=='flo':        
        fout = os.path.join(opth, 'affine_flo-' \
                + os.path.basename(imflo).split('.nii')[0]+fcomment+'.nii.gz')
    # change the file name
    os.rename(fim+'.gz', fout)

    return fout

#===============================================================================


#===============================================================================
# S P M realignment of multiple images through m-scripting (dynamic PET)
#===============================================================================
def realign_mltp_spm(
        fims,
        quality = 1.0,
        fwhm = 6,
        sep = 4,
        rtm = 1,
        interp = 2,
        graph = 0,
        outpath='',
        fcomment='',
        niicopy=False,
        niisort=False):

    '''
        fims:   has to be a list of at least two files with the first one acting
                as a reference.
    '''

    #> input folder
    inpath = os.path.dirname(fims[0])

    #> output folder
    if outpath=='':
        outpath = os.path.join(inpath, 'align')
    else:
        outpath = os.path.join(outpath, 'align')
    
    if fims[0][-3:]=='.gz' or niicopy:
        tmpth = outpath #os.path.join(outpath, 'tmp')
        rpth = tmpth
    else:
        tmpth = outpath
        rpth = inpath

    imio.create_dir(tmpth)

    #> uncompress for SPM
    fungz = []
    for f in fims:
        if f[-3:]=='.gz':
            fun = imio.nii_ugzip(f, outpath=tmpth)
        elif os.path.isfile(f) and f.endswith('nii'):
            if niicopy:
                fun = os.path.join(tmpth, os.path.basename(f).split('.nii')[0] + '_copy.nii')
                shutil.copyfile(f, fun)
            else:
                fun = f
        else:
            print 'w> omitting file/folder:', f
        fungz.append(fun)

    if niisort:
        niisrt = imio.niisort([os.path.join(tmpth,f) for f in os.listdir(tmpth) \
                if os.path.isfile(os.path.join(tmpth,f))])
        fsrt = niisrt['files']
    else:
        fsrt = fungz

    P_input = [f+',1' for f in fsrt\
        if f.endswith('nii') and f[0]!='r' and 'mean' not in f]

    #> maximal number of characters per line (for Matlab array)
    Pinmx = max([len(f) for f in P_input])

    #> equalise the input char array
    Pineq = [f.ljust(Pinmx) for f in P_input]

    #---------------------------------------------------------------------------
    #> MATLAB realign flags for SPM
    flgs = []
    pw = '\'\''

    flgs.append('flgs.quality = '+str(quality))
    flgs.append('flgs.fwhm = '+str(fwhm))
    flgs.append('flgs.sep = '+str(sep))
    flgs.append('flgs.rtm = '+str(rtm))
    flgs.append('flgs.pw  = '+str(pw))
    flgs.append('flgs.interp = '+str(interp))
    flgs.append('flgs.graphics = '+str(graph))
    flgs.append('flgs.wrap = [0 0 0]')

    flgs.append('disp(flgs)')
    #---------------------------------------------------------------------------

    fscript = os.path.join(outpath,'pyauto_script_realign.m')

    with open(fscript, 'w') as f:
        f.write('% AUTOMATICALLY GENERATED MATLAB SCRIPT FOR SPM PET REALIGNMENT\n\n')

        f.write('%> the following flags for PET image alignment are used:\n')
        f.write("disp('m> SPM realign flags:');\n")
        for item in flgs:
            f.write("%s;\n" % item)

        f.write('\n%> the following PET images will be aligned:\n')
        f.write("disp('m> PET images for SPM realignment:');\n")

        f.write('P{1,1} = [...\n')
        for item in Pineq:
            f.write("'%s'\n" % item)
        f.write('];\n\n')
        f.write('disp(P{1,1});\n')

        f.write('spm_realign(P, flgs);')


    cmd = ['matlab', '-nodisplay', '-nodesktop', '-r',  'run('+'\''+fscript+'\''+'); exit']
    call(cmd)

    fres = glob.glob(os.path.join(rpth, 'rp*.txt'))[0]
    res = np.loadtxt(fres)

    return {'fout':fres, 'outarray': res, 'P':Pineq, 'fims':fsrt}

    

#===============================================================================
# S P M resampling of multiple images through m-scripting (dynamic PET)
#===============================================================================

def resample_mltp_spm(
        fims,
        ftr,
        interp=1.,
        which=1,
        mask=0,
        mean=0,
        graph=0,
        niisort=False,
        prefix='r_',
        outpath='',
        fcomment='',
        pickname='ref',
        copy_input=False,
        del_in_uncmpr=False,
        del_out_uncmpr=False):

    '''
        fims:   has to be a list of at least two files with the first one acting
                as a reference.
    '''

    if not isinstance(fims, list) and not isinstance(fims[0], basestring):
        raise ValueError('e> unrecognised list of input images')

    if not os.path.isfile(ftr):
        raise IOError('e> cannot open the file with translations and rotations')

    #> output path
    if outpath=='':
        opth = os.path.dirname(fims[0])
    else:
        opth = outpath

    imio.create_dir(opth)

    #> working file names (not touching the original ones)
    _fims = []

    #> decompress if necessary
    for f in fims:
        if not os.path.isfile(f) and not (f.endswith('nii') or f.endswith('nii.gz')):
            raise IOError('e> could not open file:', f)

        if f[-3:]=='.gz':
            fugz = imio.nii_ugzip(f, outpath=os.path.join(opth,'copy'))
        elif copy_input:
            fnm = os.path.basename(f).split('.nii')[0] + '_copy.nii'
            fugz = os.path.join(opth, 'copy', fnm)
            shutil.copyfile(f, fugz)
        else:
            fugz = f

        _fims.append(fugz)

    if niisort:
        niisrt = imio.niisort(_fims)
        _fims = niisrt['files']


    #> maximal number of characters per line (for Matlab array)
    Pinmx = max([len(f) for f in _fims])

    #> equalise the input char array
    Pineq = [f.ljust(Pinmx) for f in _fims]

    #---------------------------------------------------------------------------
    #> SPM reslicing (resampling) flags
    flgs = []

    flgs.append('flgs.mask = '+str(mask))
    flgs.append('flgs.mean = '+str(mean))

    flgs.append('flgs.which = '+str(which))
    flgs.append('flgs.interp = '+str(interp))
    flgs.append('flgs.graphics = '+str(graph))
    flgs.append('flgs.wrap = [0 0 0]')
    flgs.append('flgs.prefix = '+"'"+prefix+"'")

    flgs.append('disp(flgs)')
    #---------------------------------------------------------------------------

    fsrpt = os.path.join(opth,'pyauto_script_resampling.m')
    with open(fsrpt, 'w') as f:
        f.write('% AUTOMATICALLY GENERATED MATLAB SCRIPT FOR SPM RESAMPLING PET IMAGES\n\n')

        f.write('%> the following flags for image reslicing are used:\n')
        f.write("disp('m> SPM reslicing flags:');\n")
        for item in flgs:
            f.write("%s;\n" % item)

        f.write('\n%> the following PET images will be aligned:\n')
        f.write("disp('m> PET images for SPM reslicing:');\n")

        f.write('P{1,1} = [...\n')
        for item in Pineq:
            f.write("'%s'\n" % item)
        f.write('];\n\n')
        f.write('disp(P{1,1});\n')

        # f.write('\n%> the following PET images will be aligned using the translations and rotations in X:\n')
        # f.write("X = dlmread('"+ftr+"');\n")
        # f.write('for fi = 2:'+str(len(fims))+'\n')
        # f.write("    VF = strcat(P{1,1}(fi,:),',1');\n")
        # f.write('    M_ = zeros(4,4);\n')
        # f.write('    M_(:,:) = spm_get_space(VF);\n')
        # f.write('    M = spm_matrix(X(fi,:));\n')
        # f.write('    spm_get_space(VF, M\M_(:,:));\n')
        # f.write('end\n\n')

        f.write('spm_reslice(P, flgs);\n')

    cmd = ['matlab', '-nodisplay', '-nodesktop', '-r',  'run('+'\''+fsrpt+'\''+'); exit']
    call(cmd)

    if del_in_uncmpr:
        for fpth in _fims:
            os.remove(fpth)


#===============================================================================
# V I N C I  registration and resampling
#===============================================================================

def coreg_vinci(
        fref,
        fflo,
        vc = '',
        con = '',
        vincipy_path = '',
        scheme_xml = '',
        outpath = '',
        fname_aff = '',
        pickname = 'ref',
        fcomment = '',
        flo_colourmap = 'Green',
        close_vinci = False,
        close_buff = True,
        cleanup = True,
        save_res = False,
        ):

    if scheme_xml=='':
        raise IOError('e> the Vinci schema *.xml file is not provided. \n \
                i> please add the schema file in the call: scheme_xml=...')

    #---------------------------------------------------------------------------
    #> output path
    if outpath=='' and fname_aff!='' and '/' in fname_aff:
        opth = os.path.dirname(fname_aff)
        if opth=='':
            opth = os.path.dirname(fflo)
        fname_aff = os.path.basename(fname_aff)

    elif outpath=='':
        opth = os.path.dirname(fflo)
    else:
        opth = outpath
    imio.create_dir(opth)

    imio.create_dir(os.path.join(opth, 'affine-vinci'))

    #> output floating and affine file names
    if fname_aff=='':
        if pickname=='ref':

            faff = os.path.join(
                    opth,
                    'affine-vinci',
                    'affine-ref-' \
                    +os.path.basename(fref).split('.nii')[0]+fcomment+'.xml')

            fout = os.path.join(
                    opth,
                    'affine-vinci',
                    'affine-ref-' \
                    +os.path.basename(fref).split('.nii')[0]+fcomment+'.nii.gz')

        else:
            faff = os.path.join(
                    opth,
                    'affine-vinci',
                    'affine-flo-' \
                    +os.path.basename(fflo).split('.nii')[0]+fcomment+'.xml')

            fout = os.path.join(
                    opth,
                    'affine-vinci',
                    'affine-flo-' \
                    +os.path.basename(fflo).split('.nii')[0]+fcomment+'.nii.gz')

    else:

        #> add '.xml' extension if not in the affine output file name
        if not fname_aff.endswith('.xml'):
            fname_aff += '.xml'

        faff = os.path.join(
                opth,
                'affine-vinci',
                fname_aff)

        fout = os.path.join(
                opth,
                fname_aff.split('.')[0]+'.nii.gz')
    #---------------------------------------------------------------------------

    if vincipy_path=='':
        try:
            import resources
            vincipy_path = resources.VINCIPATH
        except:
            raise NameError('e> could not import resources \
                    or find variable VINCIPATH in resources.py')

    sys.path.append(vincipy_path)

    try:
        from VinciPy import Vinci_Bin, Vinci_Connect, Vinci_Core, Vinci_XML, Vinci_ImageT
    except ImportError:
        raise ImportError('e> could not import Vinci:\n \
                check the variable VINCIPATH (path to Vinci) in resources.py')

    #> start Vinci core engine if it is not running/given
    if vc=='' or con=='':
        #> adopted from the Vinci's example RunMMMJob.py
        bin = Vinci_Bin.Vinci_Bin()
        con = Vinci_Connect.Vinci_Connect(bin)
        vinci_binpath = con.StartMyVinci()

        vc = Vinci_Core.Vinci_CoreCalc(con)
        vc.StdProject()

    #> read the registration schema file 
    f = open(scheme_xml, 'rb')
    reg_scheme = f.read()
    f.close()

    #> check scheme data file and remove possible XML comments at its beginning
    root = Vinci_XML.ElementTree.fromstring(reg_scheme)
    if root.tag != "MMM":
        sys.exit("scheme data file %s does not contain tag MMM\n"%scheme_xml)
    reg_scheme = Vinci_XML.ElementTree.tostring(root)
    # bytes -> str
    reg_scheme = reg_scheme.decode("utf-8")

    #> pick reference and floating images
    ref = Vinci_ImageT.newTemporary(vc, szFileName=fref)
    rsl = Vinci_ImageT.newTemporary(vc, szFileName=fflo)

    #> set the colour map for the floating image
    rsl.setColorSettings(CTable = flo_colourmap)

    #---------------------------------------------------------------------------
    #> perform the registration
    rsl.alignToRef(ref, reg_scheme, szRegistrationSummaryFile=faff)
    #---------------------------------------------------------------------------

    #> find the weird folder name created and to be disposed of.
    if cleanup:
        #> get the folder name (going to rubbish)
        del_dir = os.path.normpath(faff).split(os.sep)[1]
        fflo_dir = os.path.split(fflo)[0]
        del_dir = os.path.join(
            os.path.split(fflo)[0],
            os.path.normpath(faff).split(os.sep)[1])
        if os.path.isdir(del_dir):
            shutil.rmtree(del_dir)



    #> save the registered image
    if save_res:
        rsl.saveYourselfAs(bUseOffsetRotation=True, szFullFileName=fout)

    #> close image buffers for reference and floating
    if close_buff:
        rsl.killYourself()
        ref.killYourself()

    if close_vinci: con.CloseVinci(True)
    #else:          con.CloseSockets()


    return {'faff':faff, 'fim':fout, 'vinci_con':con, 'vinci_vc':vc}




#-------------------------------------------------------------------------------
# VINCI RESAMPLE
#-------------------------------------------------------------------------------

def resample_vinci(
        fref,
        fflo,
        faff,
        intrp = 0,
        fimout = '',
        fcomment = '',
        outpath = '',
        pickname = 'ref',
        vc = '',
        con = '',
        vincipy_path = '',
        atlas_resample = False,
        atlas_ref_make = False,
        atlas_ref_del = False,
        close_vinci = False,
        close_buff = True,
        ):
    ''' Resample the floating image <fflo> into the geometry of <fref>,
        using the Vinci transformation output <faff> (an *.xml file).
        Output the NIfTI file path of the resampled/resliced image.
    '''

    #---------------------------------------------------------------------------
    #> output path
    if outpath=='' and fimout!='' and '/' in fimout:
        opth = os.path.dirname(fimout)
        if opth=='':
            opth = os.path.dirname(fflo)
        fimout = os.path.basename(fimout)

    elif outpath=='':
        opth = os.path.dirname(fflo)
    else:
        opth = outpath
    imio.create_dir(opth)

    #> output floating and affine file names
    if fimout=='':
        if pickname=='ref':
            fout = os.path.join(
                    opth,
                    'affine-ref-' \
                    +os.path.basename(fref).split('.nii')[0]+fcomment+'.nii.gz')
        else:
            fout = os.path.join(
                    opth,
                    'affine-flo-' \
                    +os.path.basename(fflo).split('.nii')[0]+fcomment+'.nii.gz')
    else:
        fout = os.path.join(
                opth,
                fimout.split('.nii')[0]+'.nii.gz')
    #---------------------------------------------------------------------------


    if vincipy_path=='':
        try:
            import resources
            vincipy_path = resources.VINCIPATH
        except:
            raise NameError('e> could not import resources \
                    or find variable VINCIPATH in resources.py')

    sys.path.append(vincipy_path)

    try:
        from VinciPy import Vinci_Bin, Vinci_Connect, Vinci_Core, Vinci_XML, Vinci_ImageT
    except ImportError:
        raise ImportError('e> could not import Vinci:\n \
                check the variable VINCIPATH (path to Vinci) in resources.py')


    #---------------------------------------------------------------------------
    #> start Vinci core engine if it is not running/given
    if vc=='' or con=='':
        #> adopted from the Vinci's example RunMMMJob.py
        bin = Vinci_Bin.Vinci_Bin()
        con = Vinci_Connect.Vinci_Connect(bin)
        vinci_binpath = con.StartMyVinci()

        vc = Vinci_Core.Vinci_CoreCalc(con)
        vc.StdProject()
    #---------------------------------------------------------------------------

    if int(intrp)==0:
        interp_nn = True
    else:
        interp_nn = False

    #---------------------------------------------------------------------------
    #> change the reference image to be loaded as atlas
    if atlas_resample and atlas_ref_make:

        #> output file name for the extra reference image
        fextref = os.path.join(opth, 'reference-as-atlas.nii.gz')

        prc.nii_modify(fref, fimout=fextref, voxel_range=[0., 255.])
        
        ref = Vinci_ImageT.newTemporary(
                vc,
                szFileName = fextref,
                bLoadAsAtlas = atlas_resample)

        if atlas_ref_del:
            os.remove(fextref)

    else:
        ref = Vinci_ImageT.newTemporary(
                vc,
                szFileName = fref,
                bLoadAsAtlas = atlas_resample)
    #---------------------------------------------------------------------------

    #---------------------------------------------------------------------------
    #> reapply rigid transformation to new image


    flo = Vinci_ImageT.newTemporary(
            vc,
            szFileName = fflo,
            bLoadAsAtlas = atlas_resample)

    rsl = flo.reapplyMMMTransform(
            faff,
            refImage = ref,
            IsComputed=True)

    rsl.saveYourselfAs(
            bUseOffsetRotation = True,
            bUseNextNeighbourInterp = interp_nn,
            szFullFileName = fout)
    #---------------------------------------------------------------------------    


    #> close image buffers for reference and floating
    if close_buff:
        ref.killYourself()
        flo.killYourself()
        rsl.killYourself()
        


    if close_vinci: con.CloseVinci(True)

    return fout
    #return {'fim':fout, 'vinci_con':con, 'vinci_vc':vc}





#-------------------------------------------------------------------------------
# FSL-FLIRT REGISTRATION (AFFINE/RIGID)
#-------------------------------------------------------------------------------


def affine_fsl(
        fref,
        fflo,
        outpath='',
        fname_aff='',
        pickname = 'ref',
        fcomment='',
        executable='',
        costfun='normmi',
        dof=6,
        hstbins=256,
        verbose=True):


    if executable=='':
        if 'FSLDIR' not in os.environ:
            raise IOError('e> no FSL executable provided!')
        else:
            executable = os.path.join(os.environ['FSLDIR'], 'bin', 'flirt')

    if not os.path.isfile(executable) or not call([executable, '-version'])==0:
        raise IOError('e> no valid FSL executable provided!')

    #---------------------------------------------------------------------------
    #> output path
    if outpath=='' and fname_aff!='' and os.path.isfile(fname_aff):
        opth = os.path.dirname(fname_aff)
        if opth=='':
            opth = os.path.dirname(fflo)
        fname_aff = os.path.basename(fname_aff) 

    elif outpath=='':
        opth = os.path.dirname(fflo)
    else:
        opth = outpath
    imio.create_dir(opth)

    imio.create_dir(os.path.join(opth, 'affine-fsl'))

    #> output floating and affine file names
    if fname_aff=='':
        
        if pickname=='ref':
            faff = os.path.join(
                    opth,
                    'affine-fsl',
                    'affine-ref-' \
                    +os.path.basename(fref).split('.nii')[0]+fcomment+'.txt')

        else:
            faff = os.path.join(
                    opth,
                    'affine-fsl',
                    'affine-flo-' \
                    +os.path.basename(fflo).split('.nii')[0]+fcomment+'.txt')

    else:

        #> add '.xml' extension if not in the affine output file name
        if not fname_aff.endswith('.txt'):
            fname_aff += '.txt'

        faff = os.path.join(
                opth,
                'affine-vinci',
                fname_aff)
    #---------------------------------------------------------------------------
    
    #> command with options for FSL-FLIRT registration
    cmd = [ executable,
            '-cost', costfun,
            '-dof', str(dof),
            '-omat', faff,
            '-in', fflo,
            '-ref', fref,
            '-bins', str(hstbins)]

    #> add verbose mode if requested
    if isinstance(verbose, bool): verbose = int(verbose)
    if verbose>0: cmd.extend(['-verbose', str(verbose)])

    #> execute FSL-FLIRT command with the above options
    call(cmd)

    # convert to Numpy float
    aff = np.loadtxt(faff)

    return {'affine':aff, 'faff':faff}


#-------------------------------------------------------------------------------
# FSL RESAMPLING
#-------------------------------------------------------------------------------
def resample_fsl(
        imref,
        imflo,
        faff,
        outpath = '',
        fimout = '',
        pickname = 'ref',
        fcomment = '',
        intrp = 1,
        executable = ''):


    if executable=='':
        if 'FSLDIR' not in os.environ:
            raise IOError('e> no FSL executable provided!')
        else:
            executable = os.path.join(os.environ['FSLDIR'], 'bin', 'flirt')

    if not os.path.isfile(executable) or not call([executable, '-version'])==0:
        raise IOError('e> no valid FSL executable provided!')

    print '==================================================================='
    print ' F S L  resampling'

    #> output path
    if outpath=='' and fimout!='':
        opth = os.path.dirname(fimout)
        if opth=='':
            opth = os.path.dirname(imflo)

    elif outpath=='':
        opth = os.path.dirname(imflo)
    elif outpath!='':
        opth = outpath
    
    imio.create_dir(opth)

    #> the output naming
    if '/' in fimout:
        fout = fimout
    elif fimout!='' and not os.path.isfile(fimout):
        fout = os.path.join(opth, fimout)
    elif pickname=='ref':
        fout = os.path.join(opth, 'affine_ref-' \
                + os.path.basename(imref).split('.nii')[0]+fcomment+'.nii.gz')
    elif pickname=='flo':        
        fout = os.path.join(opth, 'affine_flo-' \
                + os.path.basename(imflo).split('.nii')[0]+fcomment+'.nii.gz')

    intrp = int(intrp)
    if isinstance(intrp, (int, long)): intrp = str(intrp)
    if intrp=='1':
        interpolation = 'trilinear'
    elif intrp=='0':
        interpolation = 'nearestneighbour'

    cmd = [ executable,
            '-in', imflo,
            '-ref', imref,
            '-out', fout,
            '-init', faff,
            '-applyxfm',
            '-interp', interpolation]
    call(cmd)

    print 'D O N E'
    print '==================================================================='

    return fout











# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
# M O T I O N
# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>

def motion_reg(
        ref,
        flo,
        fcomment = '',
        pickname = 'ref',
        outpath = '',
        fname_aff = '',
        matlab_eng='',
        rot_thresh = 1.,
        trn_thresh = 1.):

    
    if isinstance(flo, basestring):
        flolst = [flo]
    elif isinstance(flo, list) and all([os.path.isfile(f) for f in flo]):
        flolst = flo
    else:
        raise OSError('could not decode the input of floating images.')


    if matlab_eng=='':
        import matlab.engine
        eng = matlab.engine.start_matlab()
    else:
        eng = matlab_eng

    # dctout = {}
    lstout = []

    motion_rot = False
    motion_trn = False

    for i in range(len(flolst)):

        M = coreg_spm(
            ref,
            flolst[i],
            fcomment = fcomment,
            pickname = pickname,
            fname_aff = fname_aff,
            matlab_eng = eng,
            outpath = outpath,
            visual = 0,
            del_uncmpr=True)

        if any((np.abs(M['rotations'])*180/np.pi) > rot_thresh):
            print 'i> at least one rotation is above the threshold of', rot_thresh
            print '   ', M['rotations']*180/np.pi
            motion_rot = True
        if any(np.abs(M['translations']) > trn_thresh):
            print 'i> at least one translation is above the threshold of', rot_thresh
            print '   ', M['translations']
            motion_trn = True

        lstout.append({
                'regout':M, 
                'trans_mo':motion_trn,
                'rotat_mo':motion_rot,
            })

    return lstout

    #     dctout[os.path.basename(flolst[i])] = {
    #         'regout':M, 
    #         'trans_mo':motion_trn,
    #         'rotat_mo':motion_rot,
    #     }
    # return dctout






# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
# I M A G E   S I M I L A R I T Y
# <><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>

def dice_coeff(im1, im2, val=1):
    ''' Calculate Dice score for parcellation images <im1> and <im2> and ROI value <val>.
        Input images can be given as:
            1. paths to NIfTI image files or as
            2. Numpy arrays.
        The ROI value can be given as:
            1. a single integer representing one ROI out of many in the parcellation
               images (<im1> and <im2>) or as
            2. a list of integers to form a composite ROI used for the association test.
        Outputs a float number representing the Dice score.
    '''

    if isinstance(im1, basestring) and isinstance(im2, basestring) \
    and os.path.isfile(im1) and os.path.basename(im1).endswith(('nii', 'nii.gz')) \
    and os.path.isfile(im2) and os.path.basename(im2).endswith(('nii', 'nii.gz')):
        imn1 = imio.getnii(im1, output='image')
        imn2 = imio.getnii(im2, output='image')
    elif isinstance(im1, (np.ndarray, np.generic)) and isinstance(im1, (np.ndarray, np.generic)):
        imn1 = im1
        imn2 = im2
    else:
        raise TypeError('Unrecognised or Mismatched Images.')

    # a single value corresponding to one ROI
    if isinstance(val, (int, long)):
        imv1 = (imn1 == val)
        imv2 = (imn2 == val)
    # multiple values in list corresponding to a composite ROI
    elif isinstance(val, list) and all([isinstance(v, (int, long)) for v in val]):
        imv1 = (imn1==val[0])
        imv2 = (imn2==val[0])
        for v in val[1:]:
            # boolean addition to form a composite ROI
            imv1 += (imn1==v)
            imv2 += (imn2==v)
    else:
        raise TypeError('ROI Values have to be integer (single or in a list).')
        

    if imv1.shape != imv2.shape:
        raise ValueError('Shape Mismatch: Input images must have the same shape.')

    #-compute Dice coefficient
    intrsctn = np.logical_and(imv1, imv2)

    return 2. * intrsctn.sum() / (imv1.sum() + imv2.sum())




#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def dice_coeff_multiclass(im1, im2, roi2ind):
    ''' Calculate Dice score for parcellation images <im1> and <im2> and ROI value <val>.
        Input images can be given as:
            1. paths to NIfTI image files or as
            2. Numpy arrays.
        The ROI value must be given as a dictionary of lists of indexes for each ROI
        Outputs a float number representing the Dice score.
    '''

    if isinstance(im1, basestring) and isinstance(im2, basestring) \
    and os.path.isfile(im1) and os.path.basename(im1).endswith(('nii', 'nii.gz')) \
    and os.path.isfile(im2) and os.path.basename(im2).endswith(('nii', 'nii.gz')):
        imn1 = imio.getnii(im1, output='image')
        imn2 = imio.getnii(im2, output='image')
    elif isinstance(im1, (np.ndarray, np.generic)) and isinstance(im1, (np.ndarray, np.generic)):
        imn1 = im1
        imn2 = im2
    else:
        raise TypeError('Unrecognised or Mismatched Images.')

    if imn1.shape != imn2.shape:
        raise ValueError('Shape Mismatch: Input images must have the same shape.')

    out = {}
    for k in roi2ind.keys():

    	# multiple values in list corresponding to a composite ROI
        imv1 = (imn1==roi2ind[k][0])
        imv2 = (imn2==roi2ind[k][0])
        for v in roi2ind[k][1:]:
            # boolean addition to form a composite ROI
            imv1 += (imn1==v)
            imv2 += (imn2==v)

	    #-compute Dice coefficient
    	intrsctn = np.logical_and(imv1, imv2)
    	out[k] = 2. * intrsctn.sum() / (imv1.sum() + imv2.sum())

    return out





''' 
install tools for NiftyPET including:
* NiftyReg
* dcm2niix
'''
__author__      = "Pawel Markiewicz"
__copyright__   = "Copyright 2018"


import os
import sys
import multiprocessing
import platform
import shutil
import glob
from subprocess import call, Popen, PIPE
import re
import cudasetup as cs

if 'DISPLAY' in os.environ:
    from Tkinter import Tk
    from tkFileDialog import askdirectory

# -----------------------------------
#> NiftyReg git repository
repo_reg = 'https://github.com/KCL-BMEIS/niftyreg.git'
# repo_reg = 'https://cmiclab.cs.ucl.ac.uk/mmodat/niftyreg.git' #'git://git.code.sf.net/p/niftyreg/git'

#> git SHA-1 checksum for NiftyReg version used for PET/MR image registration and resampling
sha1_reg = '731a565bd42ca97ff5968adb1c06133ea72f0856'
# 'f673b7837c0824f55dedb1534b32b55bf68a2823'
#'6bf84b492050a4b9a93431209babeab9bc8f14da' 
#'62af1ca6777379316669b6934889c19863eaa708'

reg_ver = '1.5.61'
# -----------------------------------

# -----------------------------------
# dcm2niix git repository
repo_dcm = 'https://github.com/rordenlab/dcm2niix'
http_dcm_lin = 'https://github.com/rordenlab/dcm2niix/releases/download/v1.0.20181125/dcm2niix_25-Nov-2018_lnx.zip'
http_dcm_win = 'https://github.com/rordenlab/dcm2niix/releases/download/v1.0.20181125/dcm2niix_25-Nov-2018_win.zip'
http_dcm_mac = 'https://github.com/rordenlab/dcm2niix/releases/download/v1.0.20181125/dcm2niix_25-Nov-2018_mac.zip'

# git SHA-1 checksum for the version used for PET/MR
sha1_dcm = '32160d74cd266a59e81a75b655c16de27b8c7681'  
dcm_ver = '1.0.20181125'
# -----------------------------------

# PREVIOUS WORKING:
# http_dcm_lin =  'https://github.com/rordenlab/dcm2niix/releases/download/v1.0.20180622/dcm2niix_27-Jun-2018_lnx.zip'
# http_dcm_win = 'https://github.com/rordenlab/dcm2niix/releases/download/v1.0.20180622/dcm2niix_27-Jun-2018_win.zip'
# sha1_dcm =  '4b641113273d86ad73123816993092fc643ac62f'
# dcm_ver = '1.0.20180622'

# source and build folder names
dirsrc = '_src'
dirbld = '_bld'

# number of threads
ncpu = multiprocessing.cpu_count()

def query_yesno(question):
    valid = {'yes': True, 'y': True, 'ye': True,
             'no': False, 'n': False}
    prompt = ' [Y/n]: '
    while True:
        sys.stdout.write(question + prompt)
        choice = raw_input().lower()

        if choice == '':
            return True
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")

def input_path(question, default=os.path.expanduser('~')):
    while True:
        question += '['+default+']:'
        path = raw_input(question)
        if path == '':
            return default
        elif os.path.isdir(path):
            return path
        else:
            print 'e> the provided path is not valid: '+str(path)
            
#-----------------------------------------------------------------------------------------------------
def check_depends():
    print 'i> checking if [CUDA], [git] and [cmake] are installed...'

    outdct = {'cuda':True, 'git':True, 'cmake':True}

    #-check if CUDA is installed
    try:
        out = call(['nvcc', '--version'])
    except OSError:
        print 'e> CUDA (nvcc) does not seem to be installed;'
        outdct['cuda'] = False

    #-check if git is installed
    try:
        out = call(['git', '--version'])
    except OSError:
        print 'e> git does not seem to be installed;  for help, visit: https://git-scm.com/download/'
        outdct['git'] = False

    #-check if cmake is installed
    try:
        out = call(['cmake', '--version'])
    except OSError:
        print 'e> cmake does not seem to be installed;  for help, visit: https://cmake.org/download/'
        outdct['cmake'] = False

    return outdct

#--------------------------------------------------------------------
def check_version(Cnt, chcklst=['RESPATH','REGPATH','DCM2NIIX','HMUDIR']):
    ''' Check version and existence of all third-party software and input data.
        Output a dictionary with bool type of the requested bits in 'chcklst'
    '''

    # at start, assume that nothing is present yet
    output = {}
    for itm in chcklst:
        output[itm] = False

    # niftyreg reg_resample first
    if 'RESPATH' in chcklst and 'RESPATH' in Cnt:
        try:
            proc = Popen([Cnt['RESPATH'], '--version'], stdout=PIPE)
            out = proc.stdout.read()
            if reg_ver in out:
                output['RESPATH'] = True
        except OSError:
            print 'e> NiftyReg (reg_resample) either is NOT installed or is corrupt.'
    
    # niftyreg reg_aladin
    if 'REGPATH' in chcklst and 'REGPATH' in Cnt:
        try:
            proc = Popen([Cnt['REGPATH'], '--version'], stdout=PIPE)
            out = proc.stdout.read()
            if reg_ver in out:
                output['REGPATH'] = True
        except OSError:
            print 'e> NiftyReg (reg_aladin) either is NOT installed or is corrupt.'

    # dcm2niix
    if 'DCM2NIIX' in chcklst and 'DCM2NIIX' in Cnt:
        try:
            proc = Popen([Cnt['DCM2NIIX'], '-h'], stdout=PIPE)
            out = proc.stdout.read()
            ver_str = re.search('(?<=dcm2niiX version v)\d{1,2}.\d{1,2}.\d*', out)
            if ver_str and dcm_ver in ver_str.group(0):
                output['DCM2NIIX'] = True
        except OSError:
            print 'e> dcm2niix either is NOT installed or is corrupt.'

    # hdw mu-map list
    if 'HMUDIR' in chcklst and 'HMUDIR' in Cnt:
        for hi in Cnt['HMULIST']:
            if os.path.isfile(os.path.join(Cnt['HMUDIR'],hi)):
                output['HMUDIR'] = True
            else:
                output['HMUDIR'] = False
                break

    return output
#--------------------------------------------------------------------
def download_dcm2niix(Cnt, path):
    print '================================================='
    print 'i> dcm2niix will be installed directly from:'
    print '   https://github.com/rordenlab/dcm2niix/releases'
    print '================================================='

    #-create the installation folder
    if not os.path.isdir(path):
        os.mkdir(path)
    binpath = os.path.join(path, 'bin')
    if not os.path.isdir(binpath):
        os.mkdir(binpath)

    import urllib, zipfile
    if platform.system()=='Windows':
        urllib.urlretrieve(
            http_dcm_win,
            os.path.join(path, 'dcm2niix.zip')
        )
    elif platform.system()=='Linux':
        urllib.urlretrieve(
            http_dcm_lin,
            os.path.join(path, 'dcm2niix.zip')
        )
    elif platform.system()=='Darwin':
        urllib.urlretrieve(
            http_dcm_mac,
            os.path.join(path, 'dcm2niix.zip')
        )
    else:
        raise OSError('Unrecognised operating system.')

    zipf = zipfile.ZipFile(os.path.join(path, 'dcm2niix.zip'), 'r')
    zipf.extractall(os.path.join(path, 'bin'))
    zipf.close()
    Cnt['DCM2NIIX'] = glob.glob(os.path.join(os.path.join(path,'bin'), 'dcm2niix*'))[0]
    # ensure the permissions are given to the executable
    os.chmod(Cnt['DCM2NIIX'], 755)
    # update the resources.py file in ~/.niftypet
    Cnt = update_resources(Cnt)
    return Cnt


def install_tool(app, Cnt):
    ''' Install the requested software from the git 'repo'
        and check out the version given by 'sha1'.
    '''

    # get the current working directory
    cwd = os.getcwd()

    # pick the target installation folder for tools
    if 'PATHTOOLS' in Cnt and Cnt['PATHTOOLS']!='':
        path_tools = Cnt['PATHTOOLS']
    elif ('PATHTOOLS' not in Cnt or Cnt['PATHTOOLS']!=''):
        if 'DISPLAY' in os.environ and platform.system() in ['Linux', 'Windows']:
            print '>>>>> DISPLAY', os.environ['DISPLAY']
            Tk().withdraw()
            dircore = askdirectory(title='choose a place for NiftyPET tools', initialdir=os.path.expanduser('~'))
            Tk().destroy()
            # get the full (combined path)
            path_tools = os.path.join(dircore, Cnt['DIRTOOLS'])
        else:
            try:
                path_tools = input_path('Enter path for NiftyPET tools (registration, etc):')
            except:
                print 'enter the intended PATHTOOLS in resources.py located in ~/.niftypet/'
                raise ValueError('\n e> could not get the path for NiftyPET_tools \n')
        Cnt['PATHTOOLS'] = path_tools

    else:
        if platform.system() == 'Linux' :
            path_tools = os.path.join( os.path.expanduser('~'), Cnt['DIRTOOLS'] )
        elif platform.system() == 'Windows' :
            path_tools = os.path.join( os.getenv('LOCALAPPDATA'), Cnt['DIRTOOLS'] )
        else:
            print '\n=========================================================='
            print 'e> only Linux and Windows operating systems are supported!'
            print '==========================================================\n'
            raise SystemError('OS not supported!')      
        Cnt['PATHTOOLS'] = path_tools

    #create the main tools folder
    if not os.path.isdir(path_tools):
        os.mkdir(path_tools)
    # identify the specific path for the requested app
    if app=='niftyreg':
        repo = repo_reg
        sha1 = sha1_reg
        path = os.path.join(path_tools, 'niftyreg')
    elif app=='dcm2niix':
        repo = repo_dcm
        sha1 = sha1_dcm
        path = os.path.join(path_tools, 'dcm2niix')

        if not Cnt['CMPL_DCM2NIIX']:
            # avoid installing from source, instead download the full version:
            Cnt = download_dcm2niix(Cnt, path)
            return Cnt

    # Check if the source folder exists and delete it, if it does
    if os.path.isdir(path): shutil.rmtree(path)
    # Create an empty folder and enter it
    os.mkdir(path)
    os.chdir(path)

    # clone the git repository
    call(['git', 'clone', repo, dirsrc])
    os.chdir(dirsrc)
    print 'i> checking out the specific git version of the software...'
    call(['git', 'checkout', sha1])
    os.chdir('../')

    # create the building folder
    if not os.path.isdir(dirbld):
        os.mkdir(dirbld)
    # go inside the build folder
    os.chdir(dirbld)

    # run cmake with arguments
    if platform.system()=='Windows':
        cmd = ['cmake', '../'+dirsrc,
            '-DBUILD_ALL_DEP=ON',
            '-DCMAKE_INSTALL_PREFIX='+path,
            '-G', Cnt['MSVC_VRSN']]
        call(cmd)
        call(['cmake', '--build', './', '--config', 'Release', '--target', 'install'])
    elif platform.system() in ['Linux', 'Darwin']:
        cmd = ['cmake', '../'+dirsrc,
            '-DBUILD_ALL_DEP=ON',
            '-DCMAKE_INSTALL_PREFIX='+path]
        if Cnt['CMAKE_TLS_PAR']!='': cmd.append(Cnt['CMAKE_TLS_PAR'])
        call(cmd)
        call(
            ['cmake', '--build', './',
            '--config', 'Release',
            '--target', 'install',
            '--','-j', str(ncpu)]
        )

    # restore the current working directory
    os.chdir(cwd)

    if app=='niftyreg':
        try:
            Cnt['RESPATH'] = glob.glob(os.path.join(os.path.join(path,'bin'), 'reg_resample*'))[0]
            Cnt['REGPATH'] = glob.glob(os.path.join(os.path.join(path,'bin'), 'reg_aladin*'))[0]
        except IndexError:
            print 'e> NiftyReg has NOT been successfully installed.'
            raise SystemError('Failed Installation (NiftyReg)')
        # updated the file resources.py
        Cnt = update_resources(Cnt)
        # check the installation:
        chck_niftyreg = check_version(Cnt, chcklst=['RESPATH','REGPATH'])
        if not all([chck_niftyreg[k] for k in chck_niftyreg.keys()]):
            print 'e> NiftyReg has NOT been successfully installed.'
            raise SystemError('Failed Installation (NiftyReg)')

    elif app=='dcm2niix':
        try:
            Cnt['DCM2NIIX'] = glob.glob(os.path.join(os.path.join(path,'bin'), 'dcm2niix*'))[0]
        except IndexError:
            print 'e> dcm2niix has NOT been successfully installed.'
            Cnt = download_dcm2niix(Cnt, path)
        # check the installation:
        if not check_version(Cnt, chcklst=['DCM2NIIX']):
            print 'e> dcm2niix has NOT been successfully compiled from github.'
            Cnt = download_dcm2niix(Cnt, path)
    return Cnt
    

def update_resources(Cnt):
    '''Update resources.py with the paths to the new installed apps.
    '''

    # list of path names which will be saved
    key_list = ['PATHTOOLS', 'RESPATH', 'REGPATH', 'DCM2NIIX', 'HMUDIR']

    # get the local path to NiftyPET resources.py
    path_resources = cs.path_niftypet_local()
    resources_file = os.path.join(path_resources,'resources.py')

    # update resources.py
    if os.path.isfile(resources_file):
        f = open(resources_file, 'r')
        rsrc = f.read()
        f.close()
        # get the region of keeping in synch with Python
        i0 = rsrc.find('### start NiftyPET tools ###')
        i1 = rsrc.find('### end NiftyPET tools ###')
        pth_list = []
        for k in key_list:
            if k in Cnt:
                pth_list.append('\'' + Cnt[k].replace("\\","/") + '\'')
            else:
                pth_list.append('\'\'')

        # modify resources.py with the new paths
        strNew = '### start NiftyPET tools ###\n'
        for i in range(len(key_list)):
            if pth_list[i] != '\'\'':
                strNew += key_list[i]+' = '+pth_list[i] + '\n'
        rsrcNew = rsrc[:i0] + strNew + rsrc[i1:]
        f = open(resources_file, 'w')
        f.write(rsrcNew)
        f.close()

    return Cnt

#---------


# init the package folder

import resources as rs
import os


# check if the CUDA can run
if rs.CC_ARCH != '':
    
    from dinf import dev_info

    def gpuinfo(extended=False):
        ''' Run the CUDA dev_info shared library to get info about the installed GPU devices. 
        '''

        if extended:
            info = dev_info(1)
            print info
        else:
            info = dev_info(0)
        

        return info


# cdir = os.path.dirname(os.path.abspath(__file__))
# if len([f for f in os.listdir(cdir) if 'dinf.' in f])>0: